import numpy as np
from numba import jit

from ..kernels import BaseKernel


class CubicSplineKernel(BaseKernel):
    """An implementation of the Cubic Spline kernel"""

    @staticmethod
    def get_radius() -> float:
        return 2

    @staticmethod
    @jit(fastmath=True)
    def w(q: float, ndim: int):
        norm = 2 / 3 if (ndim == 1) else 10 / (7 * np.pi) if (ndim == 2) else 1 / np.pi

        return norm * ((1 - (3. / 2.) * q ** 2 + (3. / 4.) * q ** 3) * (0 <= q) * (q < 1)
                       + (1. / 4.) * (2 - q) ** 3 * (1 <= q) * (q < 2))



"""
Contains several interpolation functions which produce interpolated 2D or 1D arrays of SPH data.
"""
import numpy as np
from scipy.spatial.transform import Rotation

from ..interpolate import BaseBackend, CPUBackend, GPUBackend
from ..kernels import BaseKernel

from typing import Tuple, Union

def _snap(value: float):
    """
    Snap a number to the nearest integer.

    Return a number which is rounded to the nearest integer,
    with a 1e-4 absolute range of tolerance.

    Parameters
    ----------
    value: float
        The number to snap.

    Returns
    -------
    float: An integer (in float form) if a close integer is detected, otherwise return `value`.
    """
    if np.isclose(value, np.rint(value), atol=1e-4):
        return np.rint(value)
    else:
        return value


def _default_xy(data, x, y):
    """
    Utility function to determine the x & y columns to use during 2D interpolation.

    Parameters
    ----------
    data: SarracenDataFrame
        The particle dataset to interpolate over.
    x, y: str
        The x and y directional column labels passed to the interpolation function.

    Returns
    -------
    x, y: str
        The directional column labels to use in interpolation.
    """
    if x is None:
        x = data.xcol if not y == data.xcol else data.ycol
    if y is None:
        y = data.ycol if not x == data.ycol else data.xcol

    return x, y


def _default_xyz(data, x, y, z):
    """
    Utility function to determine the x, y and z columns to use during 3-D interpolation.

    Parameters
    ----------
    data: SarracenDataFrame
        The particle dataset to interpolate over.
    x, y, z: str
        The x, y and z directional column labels passed to the interpolation function.

    Returns
    -------
    x, y, z: str
        The directional column labels to use in interpolation.
    """
    if x is None:
        x = data.xcol if not y == data.xcol and not z == data.xcol else \
            data.ycol if not y == data.ycol and not z == data.ycol else data.zcol
    if y is None:
        y = data.ycol if not x == data.ycol and not z == data.ycol else \
            data.xcol if not x == data.xcol and not z == data.xcol else data.zcol
    if z is None:
        z = data.zcol if not x == data.zcol and not y == data.zcol else \
            data.ycol if not x == data.ycol and not y == data.ycol else data.xcol

    return x, y, z


def _snap_boundaries(data, x, y, xlim, ylim) -> Tuple[Tuple[float, float], Tuple[float, float]]:
    """
    Utility function to determine the 2-dimensional boundaries to use in 2D interpolation.

    Parameters
    ----------
    data: SarracenDataFrame
        The particle dataset to interpolate over.
    x, y: str
        The directional column labels that will be used in interpolation.
    xlim, ylim: tuple of float
        The minimum and maximum values passed to the interpolation function, in particle data space.

    Returns
    -------
    xlim, ylim: tuple of float
        The minimum and maximum values to use in interpolation, in particle data space. Defaults
        to the maximum and minimum values of `x` and `y`, snapped to the nearest integer.
    """
    # boundaries of the plot default to the maximum & minimum values of the data.
    x_min = xlim[0] if xlim is not None and xlim[0] is not None else None
    y_min = ylim[0] if ylim is not None and ylim[0] is not None else None
    x_max = xlim[1] if xlim is not None and xlim[1] is not None else None
    y_max = ylim[1] if ylim is not None and ylim[1] is not None else None

    x_min = _snap(data.loc[:, x].min()) if x_min is None else x_min
    y_min = _snap(data.loc[:, y].min()) if y_min is None else y_min
    x_max = _snap(data.loc[:, x].max()) if x_max is None else x_max
    y_max = _snap(data.loc[:, y].max()) if y_max is None else y_max

    return (x_min, x_max), (y_min, y_max)


def _set_pixels(x_pixels: int, y_pixels: int, xlim: Tuple[float, float], ylim: Tuple[float, float]) -> Tuple[int, int]:
    """
    Utility function to determine the number of pixels to interpolate over in 2D interpolation.

    Parameters
    ----------
    x_pixels, y_pixels: int
        The number of pixels in the x & y directions passed to the interpolation function.
    xlim, ylim: tuple of float
        The minimum and maximum values to use in interpolation, in particle data space.

    Returns
    -------
    x_pixels, y_pixels: int
        The number of pixels in the x & y directions to use in 2D interpolation.
    """
    # set # of pixels to maintain an aspect ratio that is the same as the underlying bounds of the data.
    if x_pixels is None and y_pixels is None:
        x_pixels = 512
    if x_pixels is None:
        x_pixels = int(np.rint(y_pixels * ((xlim[1] - xlim[0]) / (ylim[1] - ylim[0]))))
    if y_pixels is None:
        y_pixels = int(np.rint(x_pixels * ((ylim[1] - ylim[0]) / (xlim[1] - xlim[0]))))

    return x_pixels, y_pixels


def _verify_columns(data, x, y):
    """
    Verify that columns required for 2D interpolation exist in `data`.

    Parameters
    ----------
    data: SarracenDataFrame
        The particle dataset to interpolate over.
    target:
        Column label of the target variable to interpolate over.
    x, y: str
        The directional column labels that will be used in interpolation.

    Raises
    -------
    KeyError
        If `target`, `x`, `y`, mass, density, or smoothing length columns do not
        exist in `data`.
    """
    if x not in data.columns:
        raise KeyError(f"x-directional column '{x}' does not exist in the provided dataset.")
    if y not in data.columns:
        raise KeyError(f"y-directional column '{y}' does not exist in the provided dataset.")
    if data.hcol is None:
        raise KeyError("Smoothing length column does not exist in the provided dataset.")


def _check_boundaries(x_pixels: int, y_pixels: int, xlim: Tuple[float, float], ylim: Tuple[float, float]):
    """
    Verify that the pixel count and boundaries of a 2D plot describe a valid region.

    Parameters
    ----------
    x_pixels, y_pixels: int
        The number of pixels in the x & y directions passed to the interpolation function.
    xlim, ylim: tuple of float
        The minimum and maximum values to use in interpolation, in particle data space.

    Raises
    ------
    ValueError
        If `x_pixels` or `y_pixels` are less than or equal to zero, or
        if the specified `x` and `y` minimum and maximum values result in an invalid region.
    """
    if xlim[1] - xlim[0] <= 0:
        raise ValueError("`xlim` max must be greater than min!")
    if ylim[1] - ylim[0] <= 0:
        raise ValueError("`ylim` max must be greater than min!")
    if x_pixels <= 0:
        raise ValueError("`x_pixels` must be greater than zero!")
    if y_pixels <= 0:
        raise ValueError("`y_pixels` must be greater than zero!")


def _check_dimension(data, dim):
    """
    Verify that a given dataset describes data with a required number of dimensions.

    Parameters
    ----------
    data: SarracenDataFrame
        The particle dataset to interpolate over.
    dim: [2, 3]
        The number of required dimensions.

    Returns
    -------
    ValueError
        If the dataset is not `dim`-dimensional.
    """
    if data.get_dim() != dim:
        raise TypeError(f"Dataset is not {dim}-dimensional.")


def _rotate_data(data, x, y, z, rotation, origin):
    """
    Rotate vector data in a particle dataset.

    Parameters
    ----------
    data: SarracenDataFrame
        The particle dataset to interpolate over.
    x, y, z: str
        Directional column labels containing each dimension of the vector data.
    rotation: array_like or Rotation, optional
        The rotation to apply to the vector data. If defined as an array, the
        order of rotations is [z, y, x] in degrees
    origin: array_like, optional
        Point of rotation of the data, in [x, y, z] form.

    Returns
    -------
    x_data, y_data, z_data: ndarray
        The rotated x, y, and z directional data.
    """
    x_data = data[x].to_numpy()
    y_data = data[y].to_numpy()
    z_data = data[z].to_numpy()
    if rotation is not None:
        if not isinstance(rotation, Rotation):
            rotation = Rotation.from_euler('zyx', rotation, degrees=True)

        vectors = data[[x, y, z]].to_numpy()
        if origin is None:
            origin = (vectors.min(0) + vectors.max(0)) / 2

        vectors = vectors - origin
        vectors = rotation.apply(vectors)
        vectors = vectors + origin

        x_data = vectors[:, 0]
        y_data = vectors[:, 1]
        z_data = vectors[:, 2]

    return x_data, y_data, z_data


def _rotate_xyz(data, x, y, z, rotation, origin):
    """
    Rotate positional data in a particle dataset.

    Differs from _rotate_data() in that the returned data values are shuffled to ensure that
    the rotation is always applied to the global x, y, and z columns of the dataset, no matter
    the order of x, y, and z provided to this function.

    Parameters
    ----------
    data: SarracenDataFrame
        The particle dataset to interpolate over.
    x, y, z: str
        Directional column labels containing the positional column labels
    rotation: array_like or Rotation, optional
        The rotation to apply to the data. If defined as an array, the
        order of rotations is [z, y, x] in degrees
    origin: array_like, optional
        Point of rotation of the data, in [x, y, z] form.

    Returns
    -------
    x_data, y_data, z_data: ndarray
        The rotated x, y, and z directional data.
    """
    rotated_x, rotated_y, rotated_z = _rotate_data(data, data.xcol, data.ycol, data.zcol, rotation, origin)
    x_data = rotated_x if x == data.xcol else \
        rotated_y if x == data.ycol else \
            rotated_z if x == data.zcol else data[x]
    y_data = rotated_x if y == data.xcol else \
        rotated_y if y == data.ycol else \
            rotated_z if y == data.zcol else data[y]
    z_data = rotated_x if z == data.xcol else \
        rotated_y if z == data.ycol else \
            rotated_z if z == data.zcol else data[z]

    return x_data, y_data, z_data


def _get_mass(data: 'SarracenDataFrame'):
    if data._mcol == None:
        if 'mass' not in data.params:
            raise KeyError("'mass' column does not exist in this SarracenDataFrame.")
        return data.params['mass']

    return data[data._mcol].to_numpy()


def _get_density(data: 'SarracenDataFrame'):
    if data._rhocol == None:
        if not {data.hcol}.issubset(data.columns) or 'hfact' not in data.params:
            raise KeyError('Density cannot be derived from the columns in this SarracenDataFrame.')

        return ((data.params['hfact'] / data['h']) ** (data.get_dim()) * _get_mass(data)).to_numpy()

    return data[data._rhocol].to_numpy()


def _get_weight(data: 'SarracenDataFrame', target: Union[str, np.ndarray], dens_weight: bool):

    if type(target) is str:
        if target == 'rho':
            target_data = _get_density(data)
        else:
            if target not in data.columns:
                raise KeyError(f"Target column '{target}' does not exist in provided dataset.")

            target_data = data[target].to_numpy()
    elif type(target) is np.ndarray:
        target_data = target
    else:
        raise KeyError(f"Target must be of type str or ndarray. Found: '{type(target)}'")

    mass_data = _get_mass(data)
    if dens_weight:
        return target_data * mass_data
    else:
        rho_data = _get_density(data)
        return target_data * mass_data / rho_data

def _get_smoothing_lengths(data: 'SarracenDataFrame', hmin: float, x_pixels: int, y_pixels: int,
                           xlim: Tuple[float, float], ylim: Tuple[float, float]):
    """ Return the smoothing length data, imposing a minimum length if hmin is True. """
    
    if hmin:
        pix_size = (xlim[1] - xlim[0]) / x_pixels
        pix_size = np.maximum(pix_size, (ylim[1] - ylim[0]) / y_pixels)
        h_data = np.maximum(data[data._hcol].to_numpy(), 0.5 * pix_size)
    else:
        h_data = data[data._hcol].to_numpy()

    return h_data


def interpolate_2d(data: 'SarracenDataFrame', target: str, x: str = None, y: str = None, kernel: BaseKernel = None,
                   x_pixels: int = None, y_pixels: int = None, xlim: Tuple[float, float] = None,
                   ylim: Tuple[float, float] = None, exact: bool = False, backend: str = None,
                   dens_weight: bool = False, normalize: bool = True, hmin: bool = False) -> np.ndarray:
    """
    Interpolate particle data across two directional axes to a 2D grid of pixels.

    Interpolate the data within a SarracenDataFrame to a 2D grid, by interpolating the values
    of a target variable. The contributions of all particles near the interpolation area are
    summed and stored to a 2D grid.

    Parameters
    ----------
    data : SarracenDataFrame
        Particle data, in a SarracenDataFrame.
    target: str
        Column label of the target smoothing data.
    x, y: str
        Column labels of the directional axes. Defaults to the x & y columns detected in `data`.
    kernel: BaseKernel, optional
        Kernel to use for smoothing the target data. Defaults to the kernel specified in `data`.
    x_pixels, y_pixels: int, optional
        Number of pixels in the output image in the x & y directions. Default values are chosen to keep
        a consistent aspect ratio.
    xlim, ylim: tuple of float, optional
        The minimum and maximum values to use in interpolation, in particle data space. Defaults
        to the minimum and maximum values of `x` and `y`.
    exact: bool
        Whether to use exact interpolation of the data.
    backend: ['cpu', 'gpu']
        The computation backend to use when interpolating this data. Defaults to the backend specified in `data`.
    dens_weight: bool
        If True, the target will be multiplied by density. Defaults to False.
    hmin: bool
        If True, a minimum smoothing length of 0.5 * pixel size will be imposed. This ensures each particle
        contributes to at least one grid cell / pixel. Defaults to False (this may change in a future verison).

    Returns
    -------
    ndarray (2-Dimensional)
        The interpolated output image, in a 2-dimensional numpy array. Dimensions are
        structured in reverse order, where (x, y) -> [y, x].

    Raises
    -------
    ValueError
        If `x_pixels` or `y_pixels` are less than or equal to zero, or
        if the specified `x` and `y` minimum and maximum values result in an invalid region, or
        if `data` is not 2-dimensional.
    KeyError
        If `target`, `x`, `y`, mass, density, or smoothing length columns do not
        exist in `data`.
    """
    _check_dimension(data, 2)
    x, y = _default_xy(data, x, y)
    _verify_columns(data, x, y)

    xlim, ylim = _snap_boundaries(data, x, y, xlim, ylim)
    x_pixels, y_pixels = _set_pixels(x_pixels, y_pixels, xlim, ylim)
    _check_boundaries(x_pixels, y_pixels, xlim, ylim)
    w_data = _get_weight(data, target, dens_weight)

    kernel = kernel if kernel is not None else data.kernel
    backend = backend if backend is not None else data.backend

    h_data = _get_smoothing_lengths(data, hmin, x_pixels, y_pixels, xlim, ylim)

    grid = get_backend(backend). \
           interpolate_2d_render(data[x].to_numpy(), data[y].to_numpy(), w_data, h_data, kernel.w,
                              kernel.get_radius(), x_pixels, y_pixels, xlim[0], xlim[1], ylim[0], ylim[1], exact)

    if normalize:
        w_norm = _get_weight(data, np.array([1] * len(w_data)), dens_weight)
        norm_grid = get_backend(backend). \
                    interpolate_2d_render(data[x].to_numpy(), data[y].to_numpy(), w_norm, h_data,
                                          kernel.w, kernel.get_radius(), x_pixels, y_pixels, xlim[0], xlim[1], ylim[0],
                                          ylim[1], exact)
        grid = np.nan_to_num(grid / norm_grid)

    return grid


def interpolate_2d_vec(data: 'SarracenDataFrame', target_x: str, target_y: str, x: str = None, y: str = None,
                       kernel: BaseKernel = None, x_pixels: int = None, y_pixels: int = None,
                       xlim: Tuple[float, float] = None, ylim: Tuple[float, float] = None, exact: bool = False,
                       backend: str = None, dens_weight: bool = False, normalize: bool = True, hmin: bool = False):
    """
    Interpolate vector particle data across two directional axes to a 2D grid of particles.

    Interpolate the data within a SarracenDataFrame to a 2D grid, by interpolating the values
    of a target vector. The contributions of all vectors near the interpolation area are
    summed and stored to a 2D grid.

    Parameters
    ----------
    data : SarracenDataFrame
        Particle data, in a SarracenDataFrame.
    target_x, target_y: str
        Column labels of the target vector.
    x, y: str
        Column labels of the directional axes. Defaults to the x & y columns detected in `data`.
    kernel: BaseKernel, optional
        Kernel to use for smoothing the target data. Defaults to the kernel specified in `data`.
    x_pixels, y_pixels: int, optional
        Number of pixels in the output image in the x & y directions. Default values are chosen to keep
        a consistent aspect ratio.
    xlim, ylim: tuple of float, optional
        The minimum and maximum values to use in interpolation, in particle data space. Defaults
        to the minimum and maximum values of `x` and `y`.
    exact: bool
        Whether to use exact interpolation of the data.
    backend: ['cpu', 'gpu']
        The computation backend to use when interpolating this data. Defaults to the backend specified in `data`.
    dens_weight: bool
        If True, the target will be multiplied by density. Defaults to False.
    hmin: bool
        If True, a minimum smoothing length of 0.5 * pixel size will be imposed. This ensures each particle
        contributes to at least one grid cell / pixel. Defaults to False (this may change in a future verison).

    Returns
    -------
    output_x, output_y: ndarray (2-Dimensional)
        The interpolated output images, in a 2-dimensional numpy arrays. Dimensions are
        structured in reverse order, where (x, y) -> [y, x].

    Raises
    -------
    ValueError
        If `x_pixels` or `y_pixels` are less than or equal to zero, or
        if the specified `x` and `y` minimum and maximum values result in an invalid region, or
        if `data` is not 2-dimensional.
    KeyError
        If `target_x`, `target_y`, `x`, `y`, mass, density, or smoothing length columns do not
        exist in `data`.
    """
    _check_dimension(data, 2)
    x, y = _default_xy(data, x, y)
    _verify_columns(data, x, y)

    xlim, ylim = _snap_boundaries(data, x, y, xlim, ylim)
    x_pixels, y_pixels = _set_pixels(x_pixels, y_pixels, xlim, ylim)
    _check_boundaries(x_pixels, y_pixels, xlim, ylim)

    wx_data = _get_weight(data, target_x, dens_weight)
    wy_data = _get_weight(data, target_y, dens_weight)

    kernel = kernel if kernel is not None else data.kernel
    backend = backend if backend is not None else data.backend

    h_data = _get_smoothing_lengths(data, hmin, x_pixels, y_pixels, xlim, ylim)

    gridx, gridy = get_backend(backend).\
           interpolate_2d_render_vec(data[x].to_numpy(), data[y].to_numpy(), wx_data, wy_data, h_data,
                                     kernel.w, kernel.get_radius(), x_pixels, y_pixels, xlim[0], xlim[1], ylim[0],
                                     ylim[1], exact)

    if normalize:
        wx_norm = _get_weight(data, np.array([1] * len(wx_data)), dens_weight)
        wy_norm = _get_weight(data, np.array([1] * len(wy_data)), dens_weight)
        norm_gridx, norm_gridy = get_backend(backend).\
                                 interpolate_2d_render_vec(data[x].to_numpy(), data[y].to_numpy(), wx_norm, wy_norm,
                                                           h_data, kernel.w, kernel.get_radius(),
                                                           x_pixels, y_pixels, xlim[0], xlim[1], ylim[0], ylim[1],
                                                           exact)
        gridx = np.nan_to_num(gridx / norm_gridx)
        gridy = np.nan_to_num(gridy / norm_gridy)

    return (gridx, gridy)

def interpolate_2d_line(data: 'SarracenDataFrame', target: str, x: str = None, y: str = None,
                        kernel: BaseKernel = None, pixels: int = None, xlim: Tuple[float, float] = None,
                        ylim: Tuple[float, float] = None, backend: str = None, dens_weight: bool = False,
                        normalize: bool = True, hmin: bool = False) -> np.ndarray:
    """
    Interpolate particle data across two directional axes to a 1D cross-section line.

    Interpolate the data within a SarracenDataFrame to a 1D line, by interpolating the values
    of a target variable. The contributions of all particles near the specified line are
    summed and stored to a 1-dimensional array.

    Parameters
    ----------
    data : SarracenDataFrame
        Particle data, in a SarracenDataFrame.
    target: str
         Column label of the target smoothing data.
    x, y: str
        Column labels of the directional axes. Defaults to the x & y columns detected in `data`.
    kernel: BaseKernel, optional
        Kernel to use for smoothing the target data. Defaults to the kernel specified in `data`.
    pixels: int, optional
        Number of points in the resulting line plot in the x-direction.
    xlim, ylim: tuple of float, optional
        Starting and ending coordinates of the cross-section line (in particle data space). Defaults to
        the minimum and maximum values of `x` and `y`.
    backend: ['cpu', 'gpu']
        The computation backend to use when interpolating this data. Defaults to the backend specified in `data`.
    dens_weight: bool
        If True, the target will be multiplied by density. Defaults to False.
    hmin: bool
        If True, a minimum smoothing length of 0.5 * pixel size will be imposed. This ensures each particle
        contributes to at least one grid cell / pixel. Defaults to False (this may change in a future verison).

    Returns
    -------
    np.ndarray (1-Dimensional)
        The resulting interpolated output.

    Raises
    -------
    ValueError
        If `x_pixels` or `y_pixels` are less than or equal to zero, or
        if the specified `xlim` and `ylim` values are all the same (indicating a zero-length cross-section), or
        if `data` is not 2-dimensional.
    KeyError
        If `target`, `x`, `y`, mass, density, or smoothing length columns do not
        exist in `data`.
    """
    _check_dimension(data, 2)
    x, y = _default_xy(data, x, y)
    _verify_columns(data, x, y)

    w_data = _get_weight(data, target, dens_weight)

    if isinstance(xlim, float) or isinstance(xlim, int):
        xlim = xlim, xlim
    if isinstance(ylim, float) or isinstance(ylim, int):
        ylim = ylim, ylim

    xlim, ylim = _snap_boundaries(data, x, y, xlim, ylim)
    if xlim[0] == xlim[1] and ylim[0] == ylim[1]:
        raise ValueError('Zero length cross section!')

    kernel = kernel if kernel is not None else data.kernel
    backend = backend if backend is not None else data.backend
    pixels = pixels if pixels is not None else 512

    if pixels <= 0:
        raise ValueError('pixcount must be greater than zero!')

    if hmin:
        pix_size = np.sqrt((xlim[1] - xlim[0])**2 + (ylim[1] - ylim[0])**2) / pixels
        h_data = np.maximum(data[data._hcol].to_numpy(), 0.5 * pix_size)
    else:
        h_data = data[data._hcol].to_numpy()

    grid = get_backend(backend).\
           interpolate_2d_cross(data[x].to_numpy(), data[y].to_numpy(), w_data, h_data, kernel.w,
                                kernel.get_radius(), pixels, xlim[0], xlim[1], ylim[0], ylim[1])

    if normalize:
        w_norm = _get_weight(data, np.array([1] * len(w_data)), dens_weight)
        norm_grid = get_backend(backend). \
                    interpolate_2d_cross(data[x].to_numpy(), data[y].to_numpy(), w_norm, h_data,
                                         kernel.w, kernel.get_radius(), pixels, xlim[0], xlim[1], ylim[0], ylim[1])
        grid = np.nan_to_num(grid / norm_grid)

    return grid


def interpolate_3d_line(data: 'SarracenDataFrame', target: str, x: str = None, y: str = None, z: str = None,
                        kernel: BaseKernel = None, pixels: int = None, xlim: Tuple[float, float] = None,
                        ylim: Tuple[float, float] = None, zlim: Tuple[float, float] = None, backend: str = None,
                        dens_weight: bool = False, normalize: bool = True, hmin: bool = False):
    """
    Interpolate vector particle data across three directional axes to a 1D line.

    Interpolate the data within a SarracenDataFrame to a 1D line, by interpolating the values
    of a target variable. The contributions of all particles near the interpolation line are
    summed and stored to a 1D array.

    Parameters
    ----------
    data : SarracenDataFrame
            Particle data, in a SarracenDataFrame.
    target: str
        Column label of the target variable.
    x, y, z: str
        Column labels of the directional axes. Defaults to the x, y & z columns detected in `data`.
    kernel: BaseKernel, optional
        Kernel to use for smoothing the target data. Defaults to the kernel specified in `data`.
    pixels: int, optional
        Number of pixels in the output image in the x & y directions. Default values are chosen to keep
        a consistent aspect ratio.
    xlim, ylim, zlim: tuple of float, optional
        Starting and ending coordinates of the cross-section line (in particle data space). Defaults to
        the minimum and maximum values of `x`, `y`, and `z`.
    backend: ['cpu', 'gpu']
        The computation backend to use when interpolating this data. Defaults to the backend specified in `data`.
    dens_weight: bool
       If True, the target will be multiplied by density. Defaults to False.
    hmin: bool
        If True, a minimum smoothing length of 0.5 * pixel size will be imposed. This ensures each particle
        contributes to at least one grid cell / pixel. Defaults to False (this may change in a future verison).

    Returns
    -------
    output: ndarray (1-Dimensional)
        The interpolated output line.

    Raises
    -------
    ValueError
        If `pixels` are less than or equal to zero, or
        if the specified `x`, `y`, and `z` minimum and maximum values result in a zero area cross-section, or
        if `data` is not 3-dimensional.
    KeyError
        If `target`, `x`, `y`, mass, density, or smoothing length data does not exist in `data`.
    """
    _check_dimension(data, 3)
    x, y, z = _default_xyz(data, x, y, z)
    _verify_columns(data, x, y)

    w_data = _get_weight(data, target, dens_weight)

    if isinstance(xlim, float) or isinstance(xlim, int):
        xlim = xlim, xlim
    if isinstance(ylim, float) or isinstance(ylim, int):
        ylim = ylim, ylim
    if isinstance(zlim, float) or isinstance(zlim, int):
        zlim = zlim, zlim

    if zlim is None or zlim[0] is None:
        z1 = _snap(data.loc[:, z].min())
    else:
        z1 = zlim[0]
    if zlim is None or zlim[1] is None:
        z2 = _snap(data.loc[:, z].max())
    else:
        z2 = zlim[1]
    zlim = z1, z2

    xlim, ylim = _snap_boundaries(data, x, y, xlim, ylim)
    if ylim[1] == ylim[0] and xlim[1] == xlim[0] and zlim[1] == zlim[0]:
        raise ValueError('Zero length cross section!')

    kernel = kernel if kernel is not None else data.kernel
    backend = backend if backend is not None else data.backend

    if pixels <= 0:
        raise ValueError('pixcount must be greater than zero!')

    if hmin:
        pix_size = np.sqrt((xlim[1] - xlim[0])**2 + (ylim[1] - ylim[0])**2 + (zlim[1] - zlim[0])**2) / pixels
        h_data = np.maximum(data[data._hcol].to_numpy(), 0.5 * pix_size)
    else:
        h_data = data[data._hcol].to_numpy()

    grid = get_backend(backend) \
           .interpolate_3d_line(data[x].to_numpy(), data[y].to_numpy(), data[z].to_numpy(), w_data, h_data,
                                kernel.w, kernel.get_radius(), pixels, xlim[0], xlim[1], ylim[0], ylim[1], zlim[0],
                                zlim[1])

    if normalize:
        w_norm = _get_weight(data, np.array([1] * len(w_data)), dens_weight)
        norm_grid = get_backend(backend) \
                    .interpolate_3d_line(data[x].to_numpy(), data[y].to_numpy(), data[z].to_numpy(), w_norm,
                                         h_data, kernel.w, kernel.get_radius(), pixels, xlim[0], xlim[1],
                                         ylim[0], ylim[1], zlim[0], zlim[1])
        grid = np.nan_to_num(grid / norm_grid)

    return grid


def interpolate_3d_proj(data: 'SarracenDataFrame', target: str, x: str = None, y: str = None, kernel: BaseKernel = None,
                   integral_samples: int = 1000, rotation: np.ndarray = None, origin: np.ndarray = None,
                   x_pixels: int = None, y_pixels: int = None, xlim: Tuple[float, float] = None,
                   ylim: Tuple[float, float] = None, exact: bool = False, backend: str = None,
                   dens_weight: bool = None, normalize: bool = True, hmin: bool = False):
    """
    Interpolate 3D particle data to a 2D grid of pixels.

    Interpolates three-dimensional particle data in a SarracenDataFrame. The data
    is interpolated to a 2D grid of pixels, by summing contributions in columns which
    span the z-axis.

    Parameters
    ----------
    data : SarracenDataFrame
        Particle data, in a SarracenDataFrame.
    target: str
        Column label of the target smoothing data.
    x, y: str
        Column labels of the directional axes. Defaults to the x & y columns detected in `data`.
    kernel: BaseKernel, optional
        Kernel to use for smoothing the target data. Defaults to the kernel specified in `data`.
    integral_samples: int, optional
        Number of sample points to take when approximating the 2D column kernel.
    rotation: array_like or Rotation, optional
        The rotation to apply to the data before interpolation. If defined as an array, the
        order of rotations is [z, y, x] in degrees.
    origin: array_like, optional
        Point of rotation of the data, in [x, y, z] form. Defaults to the centre
        point of the bounds of the data.
    x_pixels, y_pixels: int, optional
        Number of pixels in the output image in the x & y directions. Default values are chosen to keep
        a consistent aspect ratio.
    xlim, ylim: tuple of float, optional
        The minimum and maximum values to use in interpolation, in particle data space. Defaults
        to the minimum and maximum values of `x` and `y`.
    exact: bool
        Whether to use exact interpolation of the data.
    backend: ['cpu', 'gpu']
        The computation backend to use when interpolating this data. Defaults to the backend specified in `data`.
    dens_weight: bool
        If True, the target will be multiplied by density. Defaults to True for column-integrated views,
        when the target is not density, and False for everything else.
    hmin: bool
        If True, a minimum smoothing length of 0.5 * pixel size will be imposed. This ensures each particle
        contributes to at least one grid cell / pixel. Defaults to False (this may change in a future verison).

    Returns
    -------
    ndarray (2-Dimensional)
        The interpolated output image, in a 2-dimensional numpy array. Dimensions are
        structured in reverse order, where (x, y) -> [y, x].

    Raises
    -------
    ValueError
        If `x_pixels` or `y_pixels` are less than or equal to zero, or
        if the specified `x` and `y` minimum and maximums result in an invalid region, or
        if the provided data is not 3-dimensional.
    KeyError
        If `target`, `x`, `y`, mass, density, or smoothing length columns do not
        exist in `data`.

    Notes
    -----
    Since the direction of integration is assumed to be straight across the z-axis, the z-axis column
    is not required for this type of interpolation.
    """
    _check_dimension(data, 3)
    x, y, z = _default_xyz(data, x, y, None)
    _verify_columns(data, x, y)

    if dens_weight is None:
        dens_weight = (target != 'rho')

    w_data = _get_weight(data, target, dens_weight)

    xlim, ylim = _snap_boundaries(data, x, y, xlim, ylim)
    x_pixels, y_pixels = _set_pixels(x_pixels, y_pixels, xlim, ylim)
    _check_boundaries(x_pixels, y_pixels, xlim, ylim)

    x_data, y_data, z_data = _rotate_xyz(data, x, y, z, rotation, origin)
    kernel = kernel if kernel is not None else data.kernel
    backend = backend if backend is not None else data.backend

    weight_function = kernel.get_column_kernel_func(integral_samples)

    h_data = _get_smoothing_lengths(data, hmin, x_pixels, y_pixels, xlim, ylim)

    grid = get_backend(backend) \
           .interpolate_3d_projection(x_data, y_data, z_data, w_data, h_data, weight_function,
                                  kernel.get_radius(), x_pixels, y_pixels, xlim[0], xlim[1], ylim[0], ylim[1], exact)

    if normalize:
        w_norm = _get_weight(data, np.array([1] * len(w_data)), dens_weight)
        norm_grid = get_backend(backend) \
                    .interpolate_3d_projection(x_data, y_data, z_data, w_norm, h_data, weight_function,
                                              kernel.get_radius(), x_pixels, y_pixels, xlim[0], xlim[1], ylim[0],
                                              ylim[1], exact)
        grid = np.nan_to_num(grid / norm_grid)

    return grid

def interpolate_3d_vec(data: 'SarracenDataFrame', target_x: str, target_y: str, target_z: str, x: str = None,
                       y: str = None, kernel: BaseKernel = None, integral_samples: int = 1000,
                       rotation: np.ndarray = None, origin: np.ndarray = None, x_pixels: int = None,
                       y_pixels: int = None, xlim: Tuple[float, float] = None, ylim: Tuple[float, float] = None,
                       exact: bool = False, backend: str = None, dens_weight: bool = False, normalize: bool = True,
                       hmin: bool = False):
    """
    Interpolate 3D vector particle data to a 2D grid of pixels.

    Interpolates three-dimensional vector particle data in a SarracenDataFrame. The data
    is interpolated to a 2D grid of pixels, by summing contributions in columns which
    span the z-axis.

    Parameters
    ----------
    data : SarracenDataFrame
        Particle data, in a SarracenDataFrame.
    target_x, target_y, target_z: str
        Column labels of the target vector.
    x, y: str
        Column labels of the directional axes. Defaults to the x & y columns detected in `data`.
    kernel: BaseKernel, optional
        Kernel to use for smoothing the target data. Defaults to the kernel specified in `data`.
    integral_samples: int, optional
        Number of sample points to take when approximating the 2D column kernel.
    rotation: array_like or Rotation, optional
        The rotation to apply to the data before interpolation. If defined as an array, the
        order of rotations is [z, y, x] in degrees.
    origin: array_like, optional
        Point of rotation of the data, in [x, y, z] form. Defaults to the centre
        point of the bounds of the data.
    x_pixels, y_pixels: int, optional
        Number of pixels in the output image in the x & y directions. Default values are chosen to keep
        a consistent aspect ratio.
    xlim, ylim: tuple of float, optional
        The minimum and maximum values to use in interpolation, in particle data space. Defaults
        to the minimum and maximum values of `x` and `y`.
    exact: bool
        Whether to use exact interpolation of the data.
    backend: ['cpu', 'gpu']
        The computation backend to use when interpolating this data. Defaults to the backend specified in `data`.
    dens_weight: bool
        If True, the target will be multiplied by density. Defaults to False.
    hmin: bool
        If True, a minimum smoothing length of 0.5 * pixel size will be imposed. This ensures each particle
        contributes to at least one grid cell / pixel. Defaults to False (this may change in a future verison).

    Returns
    -------
    output_x, output_y: ndarray (2-Dimensional)
        The interpolated output images. Dimensions are structured in reverse order, where (x, y) -> [y, x].

    Raises
    -------
    ValueError
        If `x_pixels` or `y_pixels` are less than or equal to zero, or
        if the specified `x` and `y` minimum and maximums result in an invalid region, or
        if the provided data is not 3-dimensional.
    KeyError
        If `target_x`, `target_y`, `x`, `y`, mass, density, or smoothing length columns do not
        exist in `data`.

    Notes
    -----
    Since the direction of integration is assumed to be straight across the z-axis, the z-axis column
    is not required for this type of interpolation.
    """

    _check_dimension(data, 3)
    x, y, z = _default_xyz(data, x, y, None)
    _verify_columns(data, x, y)

    xlim, ylim = _snap_boundaries(data, x, y, xlim, ylim)
    x_pixels, y_pixels = _set_pixels(x_pixels, y_pixels, xlim, ylim)
    _check_boundaries(x_pixels, y_pixels, xlim, ylim)

    x_data, y_data, _ = _rotate_xyz(data, x, y, z, rotation, origin)
    if target_z not in data.columns:
        raise KeyError(f"z-directional target column '{target_z}' does not exist in the provided dataset.")
    target_x_data, target_y_data, _ = _rotate_data(data, target_x, target_y, target_z, rotation, origin)

    wx_data = _get_weight(data, target_x_data, dens_weight)
    wy_data = _get_weight(data, target_y_data, dens_weight)
    h_data = _get_smoothing_lengths(data, hmin, x_pixels, y_pixels, xlim, ylim)

    kernel = kernel if kernel is not None else data.kernel
    backend = backend if backend is not None else data.backend

    weight_function = kernel.get_column_kernel_func(integral_samples)
    gridx, gridy = get_backend(backend) \
                   .interpolate_3d_projection_vec(x_data, y_data, wx_data, wy_data, h_data,
                                                  weight_function, kernel.get_radius(), x_pixels, y_pixels, xlim[0],
                                                  xlim[1], ylim[0], ylim[1], exact)
    if normalize:
        wx_norm = _get_weight(data, np.array([1] * len(wx_data)), dens_weight)
        wy_norm = _get_weight(data, np.array([1] * len(wy_data)), dens_weight)
        norm_gridx, norm_gridy = get_backend(backend) \
                    .interpolate_3d_projection_vec(x_data, y_data, wx_norm, wy_norm, h_data,
                                                  weight_function, kernel.get_radius(), x_pixels, y_pixels, xlim[0],
                                                  xlim[1], ylim[0], ylim[1], exact)
        gridx = np.nan_to_num(gridx / norm_gridx)
        gridy = np.nan_to_num(gridy / norm_gridy)

    return (gridx, gridy)



def interpolate_3d_cross(data: 'SarracenDataFrame', target: str, x: str = None, y: str = None, z: str = None,
                         z_slice: float = None, kernel: BaseKernel = None, rotation: np.ndarray = None,
                         origin: np.ndarray = None, x_pixels: int = None, y_pixels: int = None,
                         xlim: Tuple[float, float] = None, ylim: Tuple[float, float] = None, backend: str = None,
                         dens_weight: bool = False, normalize: bool = True, hmin: bool = False):
    """
    Interpolate 3D particle data to a 2D grid, using a 3D cross-section.

    Interpolates particle data in a SarracenDataFrame across three directional axes to a 2D
    grid of pixels. A cross-section is taken of the 3D data at a specific value of z, and
    the contributions of particles near the plane are interpolated to a 2D grid.

    Parameters
    ----------
    data : SarracenDataFrame
        The particle data to interpolate over.
    target: str
        The column label of the target smoothing data.
    z_slice: float
        The z-axis value to take the cross-section at. Defaults to the midpoint of the z-directional data.
    x, y, z: str
        The column labels of the directional data to interpolate over. Defaults to the x, y, and z columns
        detected in `data`.
    kernel: BaseKernel
        The kernel to use for smoothing the target data. Defaults to the kernel specified in `data`.
    rotation: array_like or Rotation, optional
        The rotation to apply to the data before interpolation. If defined as an array, the
        order of rotations is [z, y, x] in degrees.
    origin: array_like, optional
        Point of rotation of the data, in [x, y, z] form. Defaults to the centre
        point of the bounds of the data.
    x_pixels, y_pixels: int, optional
        Number of pixels in the output image in the x & y directions. Default values are chosen to keep
        a consistent aspect ratio.
    xlim, ylim: tuple of float, optional
        The minimum and maximum values to use in interpolation, in particle data space. Defaults
        to the minimum and maximum values of `x` and `y`.
    backend: ['cpu', 'gpu']
        The computation backend to use when interpolating this data. Defaults to the backend specified in `data`.
    dens_weight: bool
        If True, the target will be multiplied by density. Defaults to False.
    hmin: bool
        If True, a minimum smoothing length of 0.5 * pixel size will be imposed. This ensures each particle
        contributes to at least one grid cell / pixel. Defaults to False (this may change in a future verison).

    Returns
    -------
    ndarray (2-Dimensional)
        The interpolated output image, in a 2-dimensional numpy array. Dimensions are
        structured in reverse order, where (x, y) -> [y, x].

    Raises
    -------
    ValueError
        If `pixwidthx`, `pixwidthy`, `pixcountx`, or `pixcounty` are less than or equal to zero, or
        if the specified `x` and `y` minimum and maximums result in an invalid region, or
        if the provided data is not 3-dimensional.
    KeyError
        If `target`, `x`, `y`, `z`, mass, density, or smoothing length columns do not
        exist in `data`.
    """
    _check_dimension(data, 3)

    # x & y columns default to the variables determined by the SarracenDataFrame.
    x, y, z = _default_xyz(data, x, y, z)
    _verify_columns(data, x, y)

    # set default slice to be through the data's average z-value.
    if z_slice is None:
        z_slice = _snap(data.loc[:, z].mean())

    w_data = _get_weight(data, target, dens_weight)

    # boundaries of the plot default to the maximum & minimum values of the data.
    xlim, ylim = _snap_boundaries(data, x, y, xlim, ylim)
    x_pixels, y_pixels = _set_pixels(x_pixels, y_pixels, xlim, ylim)
    _check_boundaries(x_pixels, y_pixels, xlim, ylim)

    kernel = kernel if kernel is not None else data.kernel
    backend = backend if backend is not None else data.backend

    x_data, y_data, z_data = _rotate_xyz(data, x, y, z, rotation, origin)
    h_data = _get_smoothing_lengths(data, hmin, x_pixels, y_pixels, xlim, ylim)

    grid = get_backend(backend) \
           .interpolate_3d_cross(x_data, y_data, z_data, z_slice, w_data, h_data, kernel.w,
                              kernel.get_radius(), x_pixels, y_pixels, xlim[0], xlim[1], ylim[0], ylim[1])

    if normalize:
        w_norm = _get_weight(data, np.array([1] * len(w_data)), dens_weight)
        norm_grid = get_backend(backend) \
                    .interpolate_3d_cross(x_data, y_data, z_data, z_slice, w_norm, h_data, kernel.w,
                                          kernel.get_radius(), x_pixels, y_pixels, xlim[0], xlim[1], ylim[0], ylim[1])
        grid = np.nan_to_num(grid / norm_grid)

    return grid


def interpolate_3d_cross_vec(data: 'SarracenDataFrame', target_x: str, target_y: str, target_z: str,
                             z_slice: float = None, x: str = None, y: str = None, z: str = None,
                             kernel: BaseKernel = None, rotation: np.ndarray = None, origin: np.ndarray = None,
                             x_pixels: int = None, y_pixels: int = None, xlim: Tuple[float, float] = None,
                             ylim: Tuple[float, float] = None, backend: str = None, dens_weight: bool = False,
                             normalize: bool = True, hmin: bool = False):
    """
    Interpolate 3D vector particle data to a 2D grid, using a 3D cross-section.

    Interpolates vector particle data in a SarracenDataFrame across three directional axes to a 2D
    grid of pixels. A cross-section is taken of the 3D data at a specific value of z, and
    the contributions of vectors near the plane are interpolated to a 2D grid.

    Parameters
    ----------
    data : SarracenDataFrame
        The particle data to interpolate over.
    target_x, target_y, target_z: str
        The column labels of the target vector.
    z_slice: float
        The z-axis value to take the cross-section at. Defaults to the midpoint of the z-directional data.
    x, y, z: str
        The column labels of the directional data to interpolate over. Defaults to the x, y, and z columns
        detected in `data`.
    kernel: BaseKernel
        The kernel to use for smoothing the target data. Defaults to the kernel specified in `data`.
    rotation: array_like or Rotation, optional
        The rotation to apply to the data before interpolation. If defined as an array, the
        order of rotations is [z, y, x] in degrees.
    origin: array_like, optional
        Point of rotation of the data, in [x, y, z] form. Defaults to the centre
        point of the bounds of the data.
    x_pixels, y_pixels: int, optional
        Number of pixels in the output image in the x & y directions. Default values are chosen to keep
        a consistent aspect ratio.
    xlim, ylim: float, optional
        The minimum and maximum values to use in interpolation, in particle data space. Defaults
        to the minimum and maximum values of `x` and `y`.
    backend: ['cpu', 'gpu']
        The computation backend to use when interpolating this data. Defaults to the backend specified in `data`.
    dens_weight: bool
        If True, the target will be multiplied by density. Defaults to False.
    hmin: bool
        If True, a minimum smoothing length of 0.5 * pixel size will be imposed. This ensures each particle
        contributes to at least one grid cell / pixel. Defaults to False (this may change in a future verison).

    Returns
    -------
    output_x, output_y: ndarray (2-Dimensional)
        The interpolated output images. Dimensions are structured in reverse order, where (x, y) -> [y, x].

    Raises
    -------
    ValueError
        If `pixwidthx`, `pixwidthy`, `pixcountx`, or `pixcounty` are less than or equal to zero, or
        if the specified `x` and `y` minimum and maximums result in an invalid region, or
        if the provided data is not 3-dimensional.
    KeyError
        If `target_x`, `target_y`, `target_z`, `x`, `y`, `z`, mass, density, or smoothing length columns do not
        exist in `data`.
    """
    _check_dimension(data, 3)
    x, y, z = _default_xyz(data, x, y, z)
    _verify_columns(data, x, y)

    # set default slice to be through the data's average z-value.
    if z_slice is None:
        z_slice = _snap(data.loc[:, z].mean())

    # boundaries of the plot default to the maximum & minimum values of the data.
    xlim, ylim = _snap_boundaries(data, x, y, xlim, ylim)
    x_pixels, y_pixels = _set_pixels(x_pixels, y_pixels, xlim, ylim)
    _check_boundaries(x_pixels, y_pixels, xlim, ylim)

    x_data, y_data, z_data = _rotate_xyz(data, x, y, z, rotation, origin)
    target_x_data, target_y_data, _ = _rotate_data(data, target_x, target_y, target_z, rotation, origin)

    wx_data = _get_weight(data, target_x_data, dens_weight)
    wy_data = _get_weight(data, target_y_data, dens_weight)
    h_data = _get_smoothing_lengths(data, hmin, x_pixels, y_pixels, xlim, ylim)

    kernel = kernel if kernel is not None else data.kernel
    backend = backend if backend is not None else data.backend

    gridx, gridy = get_backend(backend) \
                   .interpolate_3d_cross_vec(x_data, y_data, z_data, z_slice, wx_data, wy_data, h_data,
                                             kernel.w, kernel.get_radius(), x_pixels, y_pixels, xlim[0], xlim[1],
                                             ylim[0], ylim[1])

    if normalize:
        wx_norm = _get_weight(data, np.array([1] * len(wx_data)), dens_weight)
        wy_norm = _get_weight(data, np.array([1] * len(wy_data)), dens_weight)
        norm_gridx, norm_gridy = get_backend(backend) \
                                 .interpolate_3d_cross_vec(x_data, y_data, z_data, z_slice, wx_norm, wy_norm,
                                                           h_data, kernel.w, kernel.get_radius(),
                                                           x_pixels, y_pixels, xlim[0], xlim[1], ylim[0], ylim[1])
        gridx = np.nan_to_num(gridx / norm_gridx)
        gridy = np.nan_to_num(gridy / norm_gridy)

    return (gridx, gridy)


def interpolate_3d_grid(data: 'SarracenDataFrame', target: str, x: str = None, y: str = None, z: str = None,
                        kernel: BaseKernel = None, rotation: np.ndarray = None, rot_origin: np.ndarray = None,
                        x_pixels: int = None, y_pixels: int = None, z_pixels: int = None,
                        xlim: Tuple[float, float] = None, ylim: Tuple[float, float] = None,
                        zlim: Tuple[float, float] = None, backend: str = None, dens_weight: bool = False,
                        normalize: bool = True, hmin: bool = False):
    """
    Interpolate 3D particle data to a 3D grid of pixels

    Interpolates particle data in a SarracenDataFrame across three directional axes to a 3D
    grid of pixels. The contributions of all particles near each 3D cell are summed and
    stored in the 3D grid.

    Parameters
    ----------
    data : SarracenDataFrame
        The particle data to interpolate over.
    target: str
        The column label of the target data.
    x, y, z: str
        The column labels of the directional data to interpolate over. Defaults to the x, y, and z columns
        detected in `data`.
    kernel: BaseKernel
        The kernel to use for smoothing the target data. Defaults to the kernel specified in `data`.
    rotation: array_like or Rotation, optional
        The rotation to apply to the data before interpolation. If defined as an array, the
        order of rotations is [z, y, x] in degrees.
    rot_origin: array_like, optional
        Point of rotation of the data, in [x, y, z] form. Defaults to the centre
        point of the bounds of the data.
    x_pixels, y_pixels, z_pixels: int, optional
        Number of pixels in the output image in the x, y & z directions. Default values are chosen to keep
        a consistent aspect ratio.
    xlim, ylim, zlim: tuple of float, optional
        The minimum and maximum values to use in interpolation, in particle data space. Defaults
        to the minimum and maximum values of `x`, `y` and `z`.
    backend: ['cpu', 'gpu']
        The computation backend to use when interpolating this data. Defaults to the backend specified in `data`.
    dens_weight: bool
        If True, the target will be multiplied by density. Defaults to False.
    hmin: bool
        If True, a minimum smoothing length of 0.5 * pixel size will be imposed. This ensures each particle
        contributes to at least one grid cell / pixel. Defaults to False (this may change in a future verison).

    Returns
    -------
    ndarray (3-Dimensional)
        The interpolated output image, in a 3-dimensional numpy array. Dimensions are structured in reverse order,
        where (x, y, z) -> [z, y, x].

    Raises
    -------
    ValueError
        If `x_pixels`, `y_pixels` or `z_pixels` are less than or equal to zero, or
        if the specified `x`, `y` and `z` minimum and maximum values result in an invalid region, or
        if `data` is not 3-dimensional.
    KeyError
        If `target`, `x`, `y`, `z`, mass, density, or smoothing length columns do not
        exist in `data`.
    """
    _check_dimension(data, 3)
    x, y, z = _default_xyz(data, x, y, z)
    _verify_columns(data, x, y)

    w_data = _get_weight(data, target, dens_weight)

    if not xlim:
        xlim = (None, None)
    if not ylim:
        ylim = (None, None)
    xlim, ylim = _snap_boundaries(data, x, y, xlim, ylim)
    zlim = zlim if zlim else (_snap(data.loc[:, z].min()), _snap(data.loc[:, z].max()))

    x_pixels, y_pixels = _set_pixels(x_pixels, y_pixels, xlim, ylim)
    z_pixels = int(np.rint(x_pixels * ((zlim[1] - zlim[0]) / (xlim[1] - xlim[0])))) if z_pixels is None else z_pixels
    _check_boundaries(x_pixels, y_pixels, xlim, ylim)
    if zlim[1] - zlim[0] <= 0:
        raise ValueError("`z_max` must be greater than `z_min`!")
    if z_pixels <= 0:
        raise ValueError("`z_pixels` must be greater than zero!")

    kernel = kernel if kernel is not None else data.kernel
    backend = backend if backend is not None else data.backend

    x_data, y_data, z_data = _rotate_xyz(data, x, y, data.zcol, rotation, rot_origin)
    h_data = _get_smoothing_lengths(data, hmin, x_pixels, y_pixels, xlim, ylim)

    grid = get_backend(backend) \
           .interpolate_3d_grid(x_data, y_data, z_data, w_data, h_data, kernel.w, kernel.get_radius(),
                                x_pixels, y_pixels, z_pixels, xlim[0], xlim[1], ylim[0], ylim[1], zlim[0], zlim[1])

    if normalize:
        w_norm = _get_weight(data, np.array([1] * len(w_data)), dens_weight)
        norm_grid = get_backend(backend) \
                    .interpolate_3d_grid(x_data, y_data, z_data, w_norm, h_data, kernel.w,
                                         kernel.get_radius(), x_pixels, y_pixels, z_pixels, xlim[0], xlim[1],
                                         ylim[0], ylim[1], zlim[0], zlim[1])
        grid = np.nan_to_num(grid / norm_grid)

    return grid


def get_backend(code: str) -> BaseBackend:
    """
    Get the interpolation backend associated with a string code.

    Parameters
    ----------
    code: str
        The code associated with the particular backend. At the moment, 'cpu' for the CPU backend, and 'gpu' for
        the GPU backend are supported.

    Returns
    -------
    CPUBackend: The backend to use for interpolation.
    """
    if code == 'cpu':
        return CPUBackend
    if code == 'gpu':
        return GPUBackend
    raise ValueError("Invalid code!")



from typing import Tuple

from numba.core.registry import CPUDispatcher
from numpy import ndarray, zeros


class BaseBackend:
    """Backend implementation of SPH interpolation functions."""

    @staticmethod
    def interpolate_2d_render(x: ndarray, y: ndarray, weight: ndarray, h: ndarray, weight_function: CPUDispatcher,
                              kernel_radius: float, x_pixels: int, y_pixels: int, x_min: float, x_max: float,
                              y_min: float, y_max: float, exact: bool) -> ndarray:
        """ Interpolate 2D particle data to a 2D grid of pixels."""
        return zeros((y_pixels, x_pixels))

    @staticmethod
    def interpolate_2d_render_vec(x: ndarray, y: ndarray, weight_x: ndarray, weight_y: ndarray, h: ndarray,
                                  weight_function: CPUDispatcher, kernel_radius: float, x_pixels: int, y_pixels: int,
                                  x_min: float, x_max: float, y_min: float, y_max: float,
                                  exact: bool) -> Tuple[ndarray, ndarray]:
        """ Interpolate 2D particle vector data to a pair of 2D grids of pixels. """
        return zeros((y_pixels, x_pixels)), zeros((y_pixels, x_pixels))

    @staticmethod
    def interpolate_2d_line(x: ndarray, y: ndarray, weight: ndarray, h: ndarray, weight_function: CPUDispatcher,
                             kernel_radius: float, pixels: int, x1: float, x2: float, y1: float, y2: float) -> ndarray:
        """ Interpolate 2D particle data to a 1D cross-sectional line. """
        return zeros(pixels)

    @staticmethod
    def interpolate_3d_line(x: ndarray, y: ndarray, z: ndarray, weight: ndarray, h: ndarray,
                            weight_function: CPUDispatcher, kernel_radius: float, pixels: int, x1: float, x2: float,
                            y1: float, y2: float, z1: float, z2: float) -> ndarray:
        """ Interpolate 3D particle data to a 1D cross-sectional line. """
        return zeros(pixels)

    @staticmethod
    def interpolate_3d_projection(x: ndarray, y: ndarray, z: ndarray, weight: ndarray, h: ndarray,
                                  weight_function: CPUDispatcher, kernel_radius: float, x_pixels: int, y_pixels: int,
                                  x_min: float, x_max: float, y_min: float, y_max: float, exact: bool) -> ndarray:
        """ Interpolate 3D particle data to a 2D grid of pixels, using column projection."""
        return zeros((y_pixels, x_pixels))

    @staticmethod
    def interpolate_3d_projection_vec(x: ndarray, y: ndarray, weight_x: ndarray, weight_y: ndarray, h: ndarray,
                                      weight_function: CPUDispatcher, kernel_radius: float, x_pixels: int,
                                      y_pixels: int, x_min: float, x_max: float, y_min: float, y_max: float,
                                      exact: bool) -> Tuple[ndarray, ndarray]:
        """ Interpolate 3D particle vector data to a pair of 2D grids of pixels, using column projection."""
        return zeros((y_pixels, x_pixels)), zeros((y_pixels, x_pixels))

    @staticmethod
    def interpolate_3d_cross(x: ndarray, y: ndarray, z: ndarray, z_slice: float, weight: ndarray, h: ndarray,
                             weight_function: CPUDispatcher, kernel_radius: float, x_pixels: int, y_pixels: int,
                             x_min: float, x_max: float, y_min: float, y_max: float) -> ndarray:
        """
        Interpolate 3D particle data to a pair of 2D grids of pixels, using a 3D cross-section at a specific z value.
        """
        return zeros((y_pixels, x_pixels))

    @staticmethod
    def interpolate_3d_cross_vec(x: ndarray, y: ndarray, z: ndarray, z_slice: float, weight_x: ndarray,
                                 weight_y: ndarray, h: ndarray, weight_function: CPUDispatcher, kernel_radius: float,
                                 x_pixels: int, y_pixels: int, x_min: float, x_max: float, y_min: float,
                                 y_max: float) -> Tuple[ndarray, ndarray]:
        """
        Interpolate 3D particle vector data to a pair of 2D grids of pixels, using a 3D cross-section at a
        specific z value.
        """
        return zeros((y_pixels, x_pixels)), zeros((y_pixels, x_pixels))

    @staticmethod
    def interpolate_3d_grid(x: ndarray, y: ndarray, z: ndarray, weight: ndarray, h: ndarray,
                            weight_function: CPUDispatcher, kernel_radius: float, x_pixels: int, y_pixels: int,
                            z_pixels: int, x_min: float, x_max: float, y_min: float, y_max: float, z_min: float,
                            z_max: float) -> ndarray:
        """
        Interpolate 3D particle data to a 3D grid of pixels.
        """
        return zeros((z_pixels, y_pixels, x_pixels))



import math
from typing import Tuple

import numpy as np
from numba import cuda
from numba.core.registry import CPUDispatcher
from numpy import ndarray

from ..interpolate.base_backend import BaseBackend
from ..kernels.cubic_spline_exact import line_int, surface_int


class GPUBackend(BaseBackend):

    @staticmethod
    def interpolate_2d_render(x: ndarray, y: ndarray, weight: ndarray, h: ndarray, weight_function: CPUDispatcher,
                              kernel_radius: float, x_pixels: int, y_pixels: int, x_min: float, x_max: float,
                              y_min: float, y_max: float, exact: bool) -> ndarray:
        if exact:
            return GPUBackend._exact_2d_render(x, y, weight, h, x_pixels, y_pixels, x_min, x_max, y_min, y_max)
        return GPUBackend._fast_2d(x, y, np.zeros(x.size), 0, weight, h, weight_function, kernel_radius, x_pixels,
                                   y_pixels, x_min, x_max, y_min, y_max, 2)

    @staticmethod
    def interpolate_2d_render_vec(x: ndarray, y: ndarray, weight_x: ndarray, weight_y: ndarray, h: ndarray,
                                  weight_function: CPUDispatcher, kernel_radius: float, x_pixels: int, y_pixels: int,
                                  x_min: float, x_max: float, y_min: float, y_max: float,
                                  exact: bool) -> Tuple[ndarray, ndarray]:
        if exact:
            return (GPUBackend._exact_2d_render(x, y, weight_x, h, x_pixels, y_pixels, x_min, x_max, y_min, y_max),
                    GPUBackend._exact_2d_render(x, y, weight_y, h, x_pixels, y_pixels, x_min, x_max, y_min, y_max))
        return (GPUBackend._fast_2d(x, y, np.zeros(x.size), 0, weight_x, h, weight_function, kernel_radius, x_pixels,
                                    y_pixels, x_min, x_max, y_min, y_max, 2),
                GPUBackend._fast_2d(x, y, np.zeros(x.size), 0, weight_y, h, weight_function, kernel_radius, x_pixels,
                                    y_pixels, x_min, x_max, y_min, y_max, 2))

    @staticmethod
    def interpolate_2d_cross(x: ndarray, y: ndarray, weight: ndarray, h: ndarray, weight_function: CPUDispatcher,
                             kernel_radius: float, pixels: int, x1: float, x2: float, y1: float, y2: float) -> ndarray:
        return GPUBackend._fast_2d_cross(x, y, weight, h, weight_function, kernel_radius, pixels, x1, x2, y1, y2)

    @staticmethod
    def interpolate_3d_line(x: ndarray, y: ndarray, z: ndarray, weight: ndarray, h: ndarray,
                            weight_function: CPUDispatcher, kernel_radius: float, pixels: int, x1: float, x2: float,
                            y1: float, y2: float, z1: float, z2: float) -> ndarray:
        return GPUBackend._fast_3d_line(x, y, z, weight, h, weight_function, kernel_radius, pixels, x1, x2, y1, y2, z1,
                                        z2)

    @staticmethod
    def interpolate_3d_projection(x: ndarray, y: ndarray, z: ndarray, weight: ndarray, h: ndarray,
                                  weight_function: CPUDispatcher, kernel_radius: float, x_pixels: int, y_pixels: int,
                                  x_min: float, x_max: float, y_min: float, y_max: float, exact: bool) -> ndarray:
        if exact:
            return GPUBackend._exact_3d_project(x, y, weight, h, x_pixels, y_pixels, x_min, x_max, y_min, y_max)
        return GPUBackend._fast_2d(x, y, np.zeros(x.size), 0, weight, h, weight_function, kernel_radius, x_pixels,
                                   y_pixels, x_min, x_max, y_min, y_max, 2)

    @staticmethod
    def interpolate_3d_projection_vec(x: ndarray, y: ndarray, weight_x: ndarray, weight_y: ndarray, h: ndarray,
                                      weight_function: CPUDispatcher, kernel_radius: float, x_pixels: int,
                                      y_pixels: int, x_min: float, x_max: float, y_min: float, y_max: float,
                                      exact: bool) -> Tuple[ndarray, ndarray]:
        if exact:
            return (GPUBackend._exact_3d_project(x, y, weight_x, h, x_pixels, y_pixels, x_min, x_max, y_min, y_max),
                    GPUBackend._exact_3d_project(x, y, weight_y, h, x_pixels, y_pixels, x_min, x_max, y_min, y_max))
        return (GPUBackend._fast_2d(x, y, np.zeros(x.size), 0, weight_x, h, weight_function, kernel_radius, x_pixels,
                                    y_pixels, x_min, x_max, y_min, y_max, 2),
                GPUBackend._fast_2d(x, y, np.zeros(y.size), 0, weight_y, h, weight_function, kernel_radius, x_pixels,
                                    y_pixels, x_min, x_max, y_min, y_max, 2))

    @staticmethod
    def interpolate_3d_cross(x: ndarray, y: ndarray, z: ndarray, z_slice: float, weight: ndarray, h: ndarray,
                             weight_function: CPUDispatcher, kernel_radius: float, x_pixels: int, y_pixels: int,
                             x_min: float, x_max: float, y_min: float, y_max: float) -> ndarray:
        return GPUBackend._fast_2d(x, y, z, z_slice, weight, h, weight_function, kernel_radius, x_pixels, y_pixels,
                                   x_min, x_max, y_min, y_max, 3)

    @staticmethod
    def interpolate_3d_cross_vec(x: ndarray, y: ndarray, z: ndarray, z_slice: float, weight_x: ndarray,
                                 weight_y: ndarray, h: ndarray, weight_function: CPUDispatcher, kernel_radius: float,
                                 x_pixels: int, y_pixels: int, x_min: float, x_max: float, y_min: float,
                                 y_max: float) -> Tuple[ndarray, ndarray]:
        return (GPUBackend._fast_2d(x, y, z, z_slice, weight_x, h, weight_function, kernel_radius, x_pixels, y_pixels,
                                    x_min, x_max, y_min, y_max, 3),
                GPUBackend._fast_2d(x, y, z, z_slice, weight_y, h, weight_function, kernel_radius, x_pixels, y_pixels,
                                    x_min, x_max, y_min, y_max, 3))

    @staticmethod
    def interpolate_3d_grid(x: ndarray, y: ndarray, z: ndarray, weight: ndarray, h: ndarray,
                            weight_function: CPUDispatcher, kernel_radius: float, x_pixels: int, y_pixels: int,
                            z_pixels: int, x_min: float, x_max: float, y_min: float, y_max: float, z_min: float,
                            z_max: float) -> ndarray:
        image = np.zeros((z_pixels, y_pixels, x_pixels))
        pixwidthz = (z_max - z_min) / z_pixels

        # todo: this should be separated from _fast_2d to reduce the unnecessary transfer of data to the graphics card.
        for z_i in np.arange(z_pixels):
            z_val = z_min + (z_i + 0.5) * pixwidthz
            image[z_i] = GPUBackend._fast_2d(x, y, z, z_val, weight, h, weight_function, kernel_radius, x_pixels,
                                             y_pixels, x_min, x_max, y_min, y_max, 3)

        return image

    # For the GPU, the numba code is compiled using a factory function approach. This is required
    # since a CUDA numba kernel cannot easily take weight_function as an argument.
    @staticmethod
    def _fast_2d(x_data, y_data, z_data, z_slice, w_data, h_data, weight_function, kernel_radius, x_pixels, y_pixels,
                 x_min, x_max, y_min, y_max, n_dims):
        # Underlying GPU numba-compiled code for interpolation to a 2D grid. Used in interpolation of 2D data,
        # and column integration / cross-sections of 3D data.
        @cuda.jit(fastmath=True)
        def _2d_func(z_slice, x_data, y_data, z_data, w_data, h_data, kernel_radius, x_pixels, y_pixels, x_min, x_max,
                     y_min, y_max, n_dims, image):
            pixwidthx = (x_max - x_min) / x_pixels
            pixwidthy = (y_max - y_min) / y_pixels

            i = cuda.grid(1)
            if i < len(x_data):
                if not n_dims == 2:
                    dz = np.float64(z_slice) - z_data[i]
                else:
                    dz = 0

                term = w_data[i] / h_data[i] ** n_dims

                if abs(dz) >= kernel_radius * h_data[i]:
                    return

                # determine maximum and minimum pixels that this particle contributes to
                ipixmin = round((x_data[i] - kernel_radius * h_data[i] - x_min) / pixwidthx)
                jpixmin = round((y_data[i] - kernel_radius * h_data[i] - y_min) / pixwidthy)
                ipixmax = round((x_data[i] + kernel_radius * h_data[i] - x_min) / pixwidthx)
                jpixmax = round((y_data[i] + kernel_radius * h_data[i] - y_min) / pixwidthy)

                if ipixmax < 0 or ipixmin > x_pixels or jpixmax < 0 or jpixmin > y_pixels:
                    return
                if ipixmin < 0:
                    ipixmin = 0
                if ipixmax > x_pixels:
                    ipixmax = x_pixels
                if jpixmin < 0:
                    jpixmin = 0
                if jpixmax > y_pixels:
                    jpixmax = y_pixels

                # calculate contributions to all nearby pixels
                for jpix in range(jpixmax - jpixmin):
                    for ipix in range(ipixmax - ipixmin):
                        # determine difference in the x-direction
                        xpix = x_min + ((ipix + ipixmin) + 0.5) * pixwidthx
                        dx = xpix - x_data[i]
                        dx2 = dx * dx * (1 / (h_data[i] ** 2))

                        # determine difference in the y-direction
                        ypix = y_min + ((jpix + jpixmin) + 0.5) * pixwidthy
                        dy = ypix - y_data[i]
                        dy2 = dy * dy * (1 / (h_data[i] ** 2))

                        dz2 = ((dz ** 2) * (1 / h_data[i] ** 2))

                        # calculate contributions at pixels i, j due to particle at x, y
                        q = math.sqrt(dx2 + dy2 + dz2)

                        # add contribution to image
                        if q < kernel_radius:
                            # atomic add protects the summation against race conditions.
                            wab = weight_function(q, n_dims)
                            cuda.atomic.add(image, (jpix + jpixmin, ipix + ipixmin), term * wab)

        threadsperblock = 32
        blockspergrid = (x_data.size + (threadsperblock - 1)) // threadsperblock

        # transfer relevant data to the GPU
        d_x = cuda.to_device(x_data)
        d_y = cuda.to_device(y_data)
        d_z = cuda.to_device(z_data)
        d_w = cuda.to_device(w_data)
        d_h = cuda.to_device(h_data)
        # CUDA kernels have no return values, so the image data must be
        # allocated on the device beforehand.
        d_image = cuda.to_device(np.zeros((y_pixels, x_pixels)))

        # execute the newly compiled CUDA kernel.
        _2d_func[blockspergrid, threadsperblock](z_slice, d_x, d_y, d_z, d_w, d_h, kernel_radius, x_pixels, y_pixels,
                                                 x_min, x_max, y_min, y_max, n_dims, d_image)

        return d_image.copy_to_host()

    # Underlying CPU numba-compiled code for exact interpolation of 2D data to a 2D grid.
    @staticmethod
    def _exact_2d_render(x_data, y_data, w_data, h_data, x_pixels, y_pixels, x_min, x_max, y_min, y_max):
        pixwidthx = (x_max - x_min) / x_pixels
        pixwidthy = (y_max - y_min) / y_pixels

        # Underlying GPU numba-compiled code for interpolation to a 2D grid. Used in interpolation of 2D data,
        # and column integration / cross-sections of 3D data.
        @cuda.jit
        def _2d_func(x_data, y_data, w_data, h_data, image):
            i = cuda.grid(1)
            if i < len(x_data):
                term = w_data[i] / h_data[i] ** 2

                # determine maximum and minimum pixels that this particle contributes to
                ipixmin = math.floor((x_data[i] - 2 * h_data[i] - x_min) / pixwidthx)
                jpixmin = math.floor((y_data[i] - 2 * h_data[i] - y_min) / pixwidthy)
                ipixmax = math.ceil((x_data[i] + 2 * h_data[i] - x_min) / pixwidthx)
                jpixmax = math.ceil((y_data[i] + 2 * h_data[i] - y_min) / pixwidthy)

                if ipixmax < 0 or ipixmin >= x_pixels or jpixmax < 0 or jpixmin >= y_pixels:
                    return
                if ipixmin < 0:
                    ipixmin = 0
                if ipixmax > x_pixels:
                    ipixmax = x_pixels
                if jpixmin < 0:
                    jpixmin = 0
                if jpixmax > y_pixels:
                    jpixmax = y_pixels

                denom = 1 / abs(pixwidthx * pixwidthy) * h_data[i] ** 2

                # To calculate the exact surface integral of this pixel, calculate the comprising line integrals
                # at each boundary of the square.
                if jpixmax >= jpixmin:
                    ypix = y_min + (jpixmin + 0.5) * pixwidthy
                    dy = ypix - y_data[i]

                    for ipix in range(ipixmin, ipixmax):
                        xpix = x_min + (ipix + 0.5) * pixwidthx
                        dx = xpix - x_data[i]

                        # Top Boundary
                        r0 = 0.5 * pixwidthy - dy
                        d1 = 0.5 * pixwidthx + dx
                        d2 = 0.5 * pixwidthx - dx
                        pixint = line_int(r0, d1, d2, h_data[i])
                        wab = pixint * denom

                        cuda.atomic.add(image, (jpixmin, ipix), term * wab)

                if ipixmax >= ipixmin:
                    xpix = x_min + (ipixmin + 0.5) * pixwidthx
                    dx = xpix - x_data[i]

                    for jpix in range(jpixmin, jpixmax):
                        ypix = y_min + (jpix + 0.5) * pixwidthy
                        dy = ypix - y_data[i]

                        # Left Boundary
                        r0 = 0.5 * pixwidthx - dx
                        d1 = 0.5 * pixwidthy - dy
                        d2 = 0.5 * pixwidthy + dy
                        pixint = line_int(r0, d1, d2, h_data[i])
                        wab = pixint * denom

                        cuda.atomic.add(image, (jpix, ipixmin), term * wab)

                for jpix in range(jpixmin, jpixmax):
                    ypix = y_min + (jpix + 0.5) * pixwidthy
                    dy = ypix - y_data[i]

                    for ipix in range(ipixmin, ipixmax):
                        xpix = x_min + (ipix + 0.5) * pixwidthx
                        dx = xpix - x_data[i]

                        # Bottom boundaries
                        r0 = 0.5 * pixwidthy + dy
                        d1 = 0.5 * pixwidthx - dx
                        d2 = 0.5 * pixwidthx + dx
                        pixint = line_int(r0, d1, d2, h_data[i])
                        wab = pixint * denom

                        # The negative value of the bottom boundary is equal to the value of the top boundary of the
                        # pixel below this pixel.
                        cuda.atomic.add(image, (jpix, ipix), term * wab)
                        if jpix < jpixmax - 1:
                            cuda.atomic.sub(image, (jpix + 1, ipix), term * wab)

                        # Right Boundaries
                        r0 = 0.5 * pixwidthx + dx
                        d1 = 0.5 * pixwidthy + dy
                        d2 = 0.5 * pixwidthy - dy
                        pixint = line_int(r0, d1, d2, h_data[i])
                        wab = pixint * denom

                        cuda.atomic.add(image, (jpix, ipix), term * wab)

                        # The negative value of the right boundary is equal to the value of the left boundary of the
                        # pixel to the right of this pixel.
                        if ipix < ipixmax - 1:
                            cuda.atomic.sub(image, (jpix, ipix + 1), term * wab)

        threadsperblock = 32
        blockspergrid = (x_data.size + (threadsperblock - 1)) // threadsperblock

        # transfer relevant data to the GPU
        d_x = cuda.to_device(x_data)
        d_y = cuda.to_device(y_data)
        d_w = cuda.to_device(w_data)
        d_h = cuda.to_device(h_data)
        # CUDA kernels have no return values, so the image data must be
        # allocated on the device beforehand.
        d_image = cuda.to_device(np.zeros((y_pixels, x_pixels)))

        # execute the newly compiled CUDA kernel.
        _2d_func[blockspergrid, threadsperblock](d_x, d_y, d_w, d_h, d_image)

        return d_image.copy_to_host()

    # For the GPU, the numba code is compiled using a factory function approach. This is required
    # since a CUDA numba kernel cannot easily take weight_function as an argument.
    @staticmethod
    def _fast_2d_cross(x_data, y_data, w_data, h_data, weight_function, kernel_radius, pixels, x1, x2, y1, y2):
        # determine the slope of the cross-section line
        gradient = 0
        if not x2 - x1 == 0:
            gradient = (y2 - y1) / (x2 - x1)
        yint = y2 - gradient * x2

        # determine the fraction of the line that one pixel represents
        xlength = math.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2)
        pixwidth = xlength / pixels
        xpixwidth = (x2 - x1) / pixels
        aa = 1 + gradient ** 2

        # Underlying GPU numba-compiled code for 2D->1D cross-sections
        @cuda.jit(fastmath=True)
        def _2d_func(x_data, y_data, w_data, h_data, kernel_radius, pixels, x1, x2, y1, y2, image):
            i = cuda.grid(1)
            if i < x_data.size:
                term = w_data[i] / h_data[i] ** 2

                # the intersections between the line and a particle's 'smoothing circle' are
                # found by solving a quadratic equation with the below values of a, b, and c.
                # if the determinant is negative, the particle does not contribute to the
                # cross-section, and can be removed.
                bb = 2 * gradient * (yint - y_data[i]) - 2 * x_data[i]
                cc = x_data[i] ** 2 + y_data[i] ** 2 - 2 * yint * y_data[i] + yint ** 2 - (
                            kernel_radius * h_data[i]) ** 2
                det = bb ** 2 - 4 * aa * cc

                # create a filter for particles that do not contribute to the cross-section.
                if det < 0:
                    return

                det = math.sqrt(det)

                # the starting and ending x coordinates of the lines intersections with a particle's smoothing circle.
                xstart = min(max(x1, (-bb - det) / (2 * aa)), x2)
                xend = min(max(x1, (-bb + det) / (2 * aa)), x2)

                # the start and end distances which lie within a particle's smoothing circle.
                rstart = math.sqrt((xstart - x1) ** 2 + ((gradient * xstart + yint) - y1) ** 2)
                rend = math.sqrt((xend - x1) ** 2 + (((gradient * xend + yint) - y1) ** 2))

                # the maximum and minimum pixels that each particle contributes to.
                ipixmin = min(max(0, round(rstart / pixwidth)), pixels)
                ipixmax = min(max(0, round(rend / pixwidth)), pixels)

                # iterate through all affected pixels
                for ipix in range(ipixmin, ipixmax):
                    # determine contributions to all affected pixels for this particle
                    xpix = x1 + (ipix + 0.5) * xpixwidth
                    ypix = gradient * xpix + yint
                    dy = ypix - y_data[i]
                    dx = xpix - x_data[i]

                    q2 = (dx * dx + dy * dy) * (1 / (h_data[i] * h_data[i]))
                    wab = weight_function(math.sqrt(q2), 2)

                    # add contributions to output total.
                    cuda.atomic.add(image, ipix, wab * term)

        threadsperblock = 32
        blockspergrid = (x_data.size + (threadsperblock - 1)) // threadsperblock

        # transfer relevant data to the GPU
        d_x = cuda.to_device(x_data)
        d_y = cuda.to_device(y_data)
        d_w = cuda.to_device(w_data)
        d_h = cuda.to_device(h_data)

        # CUDA kernels have no return values, so the image data must be
        # allocated on the device beforehand.
        d_image = cuda.to_device(np.zeros(pixels))

        # execute the newly compiled GPU kernel
        _2d_func[blockspergrid, threadsperblock](d_x, d_y, d_w, d_h, kernel_radius, pixels, x1, x2, y1, y2, d_image)

        return d_image.copy_to_host()

    @staticmethod
    def _fast_3d_line(x_data, y_data, z_data, w_data, h_data, weight_function, kernel_radius, pixels, x1, x2, y1, y2,
                      z1, z2):
        dx = x2 - x1
        dy = y2 - y1
        dz = z2 - z1

        length = np.sqrt(dx ** 2 + dy ** 2 + dz ** 2)
        ux, uy, uz = dx / length, dy / length, dz / length

        @cuda.jit(fastmath=True)
        def _2d_func(x_data, y_data, z_data, w_data, h_data, kernel_radius, pixels, x1, x2, y1, y2, z1, z2, image):
            i = cuda.grid(1)

            if i < x_data.size:
                delta = (ux * (x1 - x_data[i]) + uy * (y1 - y_data[i]) + uz * (z1 - z_data[i])) ** 2 \
                        - ((x1 - x_data[i]) ** 2 + (y1 - y_data[i]) ** 2 + (z1 - z_data[i]) ** 2)\
                        + (kernel_radius * h_data[i]) ** 2
                if delta < 0:
                    return

                term = w_data[i] / h_data[i] ** 3

                d1 = -(ux * (x1 - x_data[i]) + uy * (y1 - y_data[i]) + uz * (z1 - z_data[i])) - math.sqrt(delta)
                d2 = -(ux * (x1 - x_data[i]) + uy * (y1 - y_data[i]) + uz * (z1 - z_data[i])) + math.sqrt(delta)

                pixmin = min(max(0, round((d1 / length) * pixels)), pixels)
                pixmax = min(max(0, round((d2 / length) * pixels)), pixels)

                for ipix in range(pixmin, pixmax):
                    xpix = x1 + (ipix + 0.5) * (x2 - x1) / pixels
                    ypix = y1 + (ipix + 0.5) * (y2 - y1) / pixels
                    zpix = z1 + (ipix + 0.5) * (z2 - z1) / pixels

                    xdiff = xpix - x_data[i]
                    ydiff = ypix - y_data[i]
                    zdiff = zpix - z_data[i]

                    q2 = (xdiff ** 2 + ydiff ** 2 + zdiff ** 2) * (1 / (h_data[i] ** 2))
                    wab = weight_function(math.sqrt(q2), 3)

                    cuda.atomic.add(image, ipix, wab * term)

        threadsperblock = 32
        blockspergrid = (x_data.size + (threadsperblock - 1)) // threadsperblock

        # transfer relevant data to the GPU
        d_x = cuda.to_device(x_data)
        d_y = cuda.to_device(y_data)
        d_z = cuda.to_device(z_data)
        d_w = cuda.to_device(w_data)
        d_h = cuda.to_device(h_data)

        # CUDA kernels have no return values, so the image data must be
        # allocated on the device beforehand.
        d_image = cuda.to_device(np.zeros(pixels))

        # execute the newly compiled GPU kernel
        _2d_func[blockspergrid, threadsperblock](d_x, d_y, d_z, d_w, d_h, kernel_radius, pixels, x1, x2, y1, y2, z1, z2,
                                                 d_image)

        return d_image.copy_to_host()

    @staticmethod
    def _exact_3d_project(x_data, y_data, w_data, h_data, x_pixels, y_pixels, x_min, x_max, y_min, y_max):
        pixwidthx = (x_max - x_min) / x_pixels
        pixwidthy = (y_max - y_min) / y_pixels

        norm3d = 1 / np.pi

        @cuda.jit
        def _3d_func(x_data, y_data, w_data, h_data, image):
            i = cuda.grid(1)
            if i < len(x_data):
                dfac = h_data[i] ** 3 / (pixwidthx * pixwidthy * norm3d)
                term = norm3d * w_data[i] / h_data[i] ** 3

                # determine maximum and minimum pixels that this particle contributes to
                ipixmin = math.floor((x_data[i] - 2 * h_data[i] - x_min) / pixwidthx)
                jpixmin = math.floor((y_data[i] - 2 * h_data[i] - y_min) / pixwidthy)
                ipixmax = math.ceil((x_data[i] + 2 * h_data[i] - x_min) / pixwidthx)
                jpixmax = math.ceil((y_data[i] + 2 * h_data[i] - y_min) / pixwidthy)

                # The width of the z contribution of this particle.
                # = 2 * kernel_radius * h[i], where kernel_radius is 2 for the cubic spline kernel.
                pixwidthz = 4 * h_data[i]

                if ipixmax < 0 or ipixmin >= x_pixels or jpixmax < 0 or jpixmin >= y_pixels:
                    return
                if ipixmin < 0:
                    ipixmin = 0
                if ipixmax > x_pixels:
                    ipixmax = x_pixels
                if jpixmin < 0:
                    jpixmin = 0
                if jpixmax > y_pixels:
                    jpixmax = y_pixels

                for jpix in range(jpixmin, jpixmax):
                    ypix = y_min + (jpix + 0.5) * pixwidthy
                    dy = ypix - y_data[i]

                    for ipix in range(ipixmin, ipixmax):
                        xpix = x_min + (ipix + 0.5) * pixwidthx
                        dx = xpix - x_data[i]

                        q2 = (dx ** 2 + dy ** 2) / h_data[i] ** 2

                        if q2 < 4 + 3 * pixwidthx * pixwidthy / h_data[i] ** 2:
                            # Calculate the volume integral of this pixel by summing the comprising
                            # surface integrals of each surface of the cube.

                            # x-y surfaces
                            pixint = 2 * surface_int(0.5 * pixwidthz, x_data[i], y_data[i], xpix, ypix, pixwidthx,
                                                     pixwidthy, h_data[i])

                            # x-z surfaces
                            pixint += surface_int(ypix - y_data[i] + 0.5 * pixwidthy, x_data[i], 0, xpix, 0, pixwidthx,
                                                  pixwidthz, h_data[i])
                            pixint += surface_int(y_data[i] - ypix + 0.5 * pixwidthy, x_data[i], 0, xpix, 0, pixwidthx,
                                                  pixwidthz, h_data[i])

                            # y-z surfaces
                            pixint += surface_int(xpix - x_data[i] + 0.5 * pixwidthx, 0, y_data[i], 0, ypix, pixwidthz,
                                                  pixwidthy, h_data[i])
                            pixint += surface_int(x_data[i] - xpix + 0.5 * pixwidthx, 0, y_data[i], 0, ypix, pixwidthz,
                                                  pixwidthy, h_data[i])

                            wab = pixint * dfac

                            cuda.atomic.add(image, (jpix, ipix), term * wab)

        threadsperblock = 32
        blockspergrid = (x_data.size + (threadsperblock - 1)) // threadsperblock

        # transfer relevant data to the GPU
        d_x = cuda.to_device(x_data)
        d_y = cuda.to_device(y_data)
        d_w = cuda.to_device(w_data)
        d_h = cuda.to_device(h_data)
        # CUDA kernels have no return values, so the image data must be
        # allocated on the device beforehand.
        d_image = cuda.to_device(np.zeros((y_pixels, x_pixels)))

        # execute the newly compiled CUDA kernel.
        _3d_func[blockspergrid, threadsperblock](d_x, d_y, d_w, d_h, d_image)

        return d_image.copy_to_host()



"""pytest unit tests for kernel functionality."""
from pytest import approx
from scipy.integrate import quad, dblquad, tplquad
import numpy as np

from sarracen.kernels import CubicSplineKernel, QuarticSplineKernel, QuinticSplineKernel


def single_kernel(x, kernel):
    return kernel.w(np.abs(x), 1)


def double_kernel(y, x, kernel):
    # Utility function for double integrals in test_normalization
    return kernel.w(np.sqrt(x ** 2 + y ** 2), 2)


def triple_kernel(z, y, x, kernel):
    # Utility function for triple integrals in test_normalization
    return kernel.w(np.sqrt(x ** 2 + y ** 2 + z ** 2), 3)


def double_column(y, x, column_func):
    return column_func(np.sqrt(x ** 2 + y ** 2), 0)


class TestKernels:
    def test_cubicspline(self):
        kernel = CubicSplineKernel()

        # testing kernel values at q = 0
        # which should be equal to the normalization constants
        assert kernel.w(0, 1) == 2 / 3
        assert kernel.w(0, 2) == 10 / (7 * np.pi)
        assert kernel.w(0, 3) == 1 / np.pi

        # testing kernel values at q = 1
        assert kernel.w(1, 1) == 1 / 6
        assert kernel.w(1, 2) == 5 / (14 * np.pi)
        assert kernel.w(1, 3) == 1 / (4 * np.pi)

        # testing kernel values at q = 2
        assert kernel.w(2, 2) == 0
        assert kernel.w(10, 3) == 0

    def test_quarticspline(self):
        kernel = QuarticSplineKernel()

        # unlike the cubic spline, these will NOT
        # be equal to the normalization constants.
        # kernel.w(0, d) = norm_d * 230/16
        assert kernel.w(0, 1) == approx(115 / 192)
        assert kernel.w(0, 2) == approx(1380 / (1199 * np.pi))
        assert kernel.w(0, 3) == approx(23 / (32 * np.pi))

        assert kernel.w(1, 1) == approx(19 / 96)
        assert kernel.w(1, 2) == approx(456 / (1199 * np.pi))
        assert kernel.w(1, 3) == approx(19 / (80 * np.pi))

        # these are equivalent to the normalization constants
        assert kernel.w(1.5, 1) == approx(1 / 24)
        assert kernel.w(1.5, 2) == approx(96 / (1199 * np.pi))
        assert kernel.w(1.5, 3) == approx(1 / (20 * np.pi))

        assert kernel.w(2.5, 2) == 0
        assert kernel.w(10, 3) == 0

    def test_quinticspline(self):
        kernel = QuinticSplineKernel()

        # again, unlike the cubic spline, these will NOT
        # be equal to the normalization constants.
        # kernel.w(0, d) = norm_d * 66
        assert kernel.w(0, 1) == approx(11 / 20)
        assert kernel.w(0, 2) == approx(231 / (239 * np.pi))
        assert kernel.w(0, 3) == approx(11 / (20 * np.pi))

        assert kernel.w(1, 1) == approx(13 / 60)
        assert kernel.w(1, 2) == approx(91 / (239 * np.pi))
        assert kernel.w(1, 3) == approx(13 / (60 * np.pi))

        # these are equivalent to the normalization constants
        assert kernel.w(2, 1) == approx(1 / 120)
        assert kernel.w(2, 2) == approx(7 / (478 * np.pi))
        assert kernel.w(2, 3) == approx(1 / (120 * np.pi))

        assert kernel.w(3, 2) == 0
        assert kernel.w(10, 3) == 0

    def test_normalization(self):

        # Since the three integrals below are only performed in positive space, the
        # resulting normalized values will not be equal to 1, rather 1/(2^dim). This
        # value represents the proportion of space in 1,2, and 3 dimensions that
        # has all positive coordinates.

        for kernel in [CubicSplineKernel(), QuarticSplineKernel(), QuinticSplineKernel()]:
            norm = quad(single_kernel, -kernel.get_radius(), kernel.get_radius(), kernel)[0]
            assert approx(norm) == 1

        for kernel in [CubicSplineKernel(), QuarticSplineKernel(), QuinticSplineKernel()]:
            norm = dblquad(double_kernel, -kernel.get_radius(), kernel.get_radius(), -kernel.get_radius(),
                           kernel.get_radius(), [kernel])[0]
            assert approx(norm) == 1

        for kernel in [CubicSplineKernel(), QuarticSplineKernel(), QuinticSplineKernel()]:
            norm = tplquad(triple_kernel, -kernel.get_radius(), kernel.get_radius(), -kernel.get_radius(),
                           kernel.get_radius(), -kernel.get_radius(), kernel.get_radius(), [kernel])[0]
            assert approx(norm) == 1

    def test_cubic_column(self):
        kernel = CubicSplineKernel()
        column_kernel = kernel.get_column_kernel(10000)

        # at q = 0, this integral is solvable analytically
        assert np.interp(0, np.linspace(0, kernel.get_radius(), 10000), column_kernel) == approx(3 / (2 * np.pi))
        # numerically calculated values
        assert np.interp(0.5, np.linspace(0, kernel.get_radius(), 10000), column_kernel) == approx(0.33875339978)
        assert np.interp(1, np.linspace(0, kernel.get_radius(), 10000), column_kernel) == approx(0.111036060968)
        assert np.interp(1.5, np.linspace(0, kernel.get_radius(), 10000), column_kernel) == approx(0.0114423169642)
        assert np.interp(2, np.linspace(0, kernel.get_radius(), 10000), column_kernel) == 0
        assert np.interp(5, np.linspace(0, kernel.get_radius(), 10000), column_kernel) == 0

    def test_quartic_column(self):
        kernel = QuarticSplineKernel()
        column_kernel = kernel.get_column_kernel(10000)

        # at q = 0, this integral is solvable analytically
        assert np.interp(0, np.linspace(0, kernel.get_radius(), 10000), column_kernel) == approx(6 / (5 * np.pi))
        # numerically calculated values
        assert np.interp(0.5, np.linspace(0, kernel.get_radius(), 10000), column_kernel) == approx(0.288815941868)
        assert np.interp(1, np.linspace(0, kernel.get_radius(), 10000), column_kernel) == approx(0.120120735858)
        assert np.interp(1.5, np.linspace(0, kernel.get_radius(), 10000), column_kernel) == approx(0.0233911861393)
        assert np.interp(2, np.linspace(0, kernel.get_radius(), 10000), column_kernel) == approx(0.00116251851966)
        assert np.interp(2.5, np.linspace(0, kernel.get_radius(), 10000), column_kernel) == 0
        assert np.interp(5, np.linspace(0, kernel.get_radius(), 10000), column_kernel) == 0

    def test_quintic_column(self):
        kernel = QuinticSplineKernel()
        column_kernel = kernel.get_column_kernel(10000)

        # at q = 0, this integral is solvable analytically
        assert np.interp(0, np.linspace(0, kernel.get_radius(), 10000), column_kernel) == approx(1 / np.pi)
        # numerically calculated values
        assert np.interp(0.5, np.linspace(0, kernel.get_radius(), 10000), column_kernel) == approx(0.251567608959)
        assert np.interp(1, np.linspace(0, kernel.get_radius(), 10000), column_kernel) == approx(0.121333261458)
        assert np.interp(1.5, np.linspace(0, kernel.get_radius(), 10000), column_kernel) == approx(0.0328632154395)
        assert np.interp(2, np.linspace(0, kernel.get_radius(), 10000), column_kernel) == approx(0.00403036583315)
        assert np.interp(2.5, np.linspace(0, kernel.get_radius(), 10000), column_kernel) == approx(0.0000979416858548,
                                                                                                   rel=1e-4)
        assert np.interp(3, np.linspace(0, kernel.get_radius(), 10000), column_kernel) == 0
        assert np.interp(5, np.linspace(0, kernel.get_radius(), 10000), column_kernel) == 0

    def test_normalized_column(self):
        for kernel in [CubicSplineKernel(), QuarticSplineKernel(), QuinticSplineKernel()]:
            norm = dblquad(double_column, -kernel.get_radius(), kernel.get_radius(), -kernel.get_radius(),
                           kernel.get_radius(), [kernel.get_column_kernel_func(10000)])[0]
            assert approx(norm) == 1

    def test_oob(self):
        for kernel in [CubicSplineKernel(), QuarticSplineKernel(), QuinticSplineKernel()]:
            for dimensions in range(1, 3):
                assert kernel.w(-1, dimensions) == 0
                assert kernel.w(kernel.get_radius() + 1, dimensions) == 0

            column_func = kernel.get_column_kernel_func(1000)
            assert column_func(-1, 0) == column_func(0, 0)
            assert approx(column_func(kernel.get_radius() + 1, 0)) == 0



import pandas as pd
import numpy as np
from pandas import testing as tm
from sarracen.readers import read_csv


def test_get_units():
    test_units = pd.Series(['va_r?s w/ sp3ci@l ch%rs [units /$:. ]',
                 'a [a]',
                 'a',
                 '[]'])
    test_units = read_csv._get_units(test_units)

    answer_units = ['units /$:. ',
                   'a',
                   None,
                   None]
    answer_units = pd.Series(answer_units, name=0)

    tm.assert_series_equal(test_units, answer_units)



def test_get_labels():
    test_labels = pd.Series(['va_r?s w/ sp3ci@l ch%rs [units /$:. ]',
             'a [a]',
             'a',
             '[]'])
    test_labels = read_csv._get_labels(test_labels)

    answer_labels = ['va_r?s w/ sp3ci@l ch%rs',
                'a',
                'a',
                None]
    answer_labels = pd.Series(answer_labels, name=0)

    tm.assert_series_equal(test_labels, answer_labels)




from .readers.read_marisa import read_marisa
from .readers.read_csv import read_csv
from .readers.read_phantom import read_phantom

from .sarracen_dataframe import SarracenDataFrame

from .interpolate import interpolate_2d, interpolate_2d_line, interpolate_3d_proj, interpolate_3d_cross
from .render import render, streamlines, arrowplot


__version__ = "1.2.0"



from typing import Union, Callable, Tuple

from matplotlib.axes import Axes
from matplotlib.colors import Colormap
from pandas import DataFrame, Series
import numpy as np

from .render import streamlines, arrowplot, render, lineplot
from .interpolate import interpolate_2d, interpolate_3d_grid
from .kernels import CubicSplineKernel, BaseKernel

from typing import Tuple

def _copy_doc(copy_func: Callable) -> Callable:
    """Copy documentation from another function to this function."""
    def wrapper(func: Callable) -> Callable:
        func.__doc__ = copy_func.__doc__
        return func
    return wrapper


class SarracenDataFrame(DataFrame):
    """
    A SarracenDataFrame is a pandas DataFrame with support for SPH data.

    A SarracenDataFrame is a subclass of the pandas DataFrame class designed to hold SPH particle
    data. Global simulation values are stored in ``params``, which is a standard Python dictionary.

    Interpolation and rendering functionality requires (at a minimum) particle positions, smoothing
    lengths and masses. SarracenDataFrames will attempt to identify columns which hold these data.
    For uniform, constant mass particles, the particle mass can be specified in the ``params``
    dictionary.

    """

    _metadata = ['_params', '_units', '_xcol', '_ycol', '_zcol', '_hcol', '_mcol', '_rhocol', '_kernel']

    def __init__(self, data=None, params=None, *args, **kwargs):
        """
        Construct a SarracenDataFrame from a NumPy array, dictionary, DataFrame or Iterable object.

        Parameters
        ----------
        data : ndarray, Iterable, DataFrame, or dict.
            Raw particle data which is passed to the pandas DataFrame constructor. Data can be specified
            in a dictionary, NumPy array or another DataFrame.
        params : dict, optional
            Global parameters from the simulation (time, hfact, etc). If constant, uniform mass particles
            are used, then the key ``mass`` stores the particle mass (rather than specifying per particle).
        *args : tuple, optional
            Additional arguments to pass to the pandas DataFrame constructor.
        **kwargs : dict, optional
            Additional keyword arguments to pass to the pandas DataFrame constructor.

        See Also
        --------
        :func:`read_csv` : Read data from a comma separated values (csv) file.
        :func:`read_phantom` : Read data from the Phantom SPH code.

        Examples
        --------
        Constructing using a Python dictionary.

        >>> particles = {'x': [1.0, 2.0, 3.0], 'y': [2.0, 2.0, 2.0], 'h': [3.0, 3.5, 4.0]}
        >>> sdf = sarracen.SarracenDataFrame(particles)
        >>> sdf
            x     y     h
        0   1.0   2.0   3.0
        1   2.0   2.0   3.5
        2   3.0   2.0   4.0

        Constructing using a two-dimensional NumPy array.

        >>> particles = np.array([[1.0, 2.0, 3.0], [2.0, 2.0, 3.5], [3.0, 2.0, 4.0]])
        >>> sdf = sarracen.SarracenDataFrame(particles, columns=['x', 'y', 'h'])
        >>> sdf
            x     y     h
        0   1.0   2.0   3.0
        1   2.0   2.0   3.5
        2   3.0   2.0   4.0

        Constant mass particles can specify mass in the ``params`` dictionary, rather than per particle.

        >>> particles = {'x': [1.0, 2.0, 3.0], 'y': [2.0, 2.0, 2.0], 'h': [3.0, 3.5, 4.0]}
        >>> params = {'mass': 0.2, 'hfact': 1.2}
        >>> sdf = sarracen.SarracenDataFrame(particles, params)
        >>> sdf.params
        {'mass': 0.2, 'hfact': 1.2}

        """

        # call pandas DataFrame constructor
        super().__init__(data, *args, **kwargs)

        if params is None:
            params = dict()
        self._params = None
        self.params = params

        self._units = None
        self.units = Series([np.nan for _ in range(len(self.columns))])

        self._xcol, self._ycol, self._zcol, self._hcol, self._mcol, self._rhocol = None, None, None, None, None, None
        self._identify_special_columns()

        self._kernel = CubicSplineKernel()
        self._backend = 'cpu'

    @property
    def _constructor(self):
        return SarracenDataFrame

    def _identify_special_columns(self):
        """
        Identify special columns commonly used in analysis functions.

        Identify which columns in this dataset correspond to important data columns commonly used in
        analysis functions. The columns which contain x, y, and z positional values are detected and
        set to the `xcol`, `ycol`, and `zcol` values. As well, the columns containing smoothing length,
        mass, and density information are identified and set to the `hcol`, `mcol`, and `rhocol`.

        If the x or y columns cannot be found, they are set to be the first two columns by default.
        If the z, smoothing length, mass, or density columns cannot be sound, the corresponding column
        label is set to `None`.
        """
        # First look for 'x', then 'rx', and then fallback to the first column.
        if 'x' in self.columns:
            self._xcol = 'x'
        elif 'rx' in self.columns:
            self._xcol = 'rx'
        else:
            self._xcol = self.columns[0]

        # First look for 'y', then 'ry', and then fallback to the second column.
        if 'y' in self.columns:
            self._ycol = 'y'
        elif 'ry' in self.columns:
            self._ycol = 'ry'
        else:
            self._ycol = self.columns[1]

        # First look for 'z', then 'rz', and then assume that data is 2 dimensional.
        if 'z' in self.columns:
            self._zcol = 'z'
        elif 'rz' in self.columns:
            self._zcol = 'rz'

        # Look for the keyword 'h' in the data.
        if 'h' in self.columns:
            self._hcol = 'h'

        # Look for the keyword 'm' or 'mass' in the data.
        if 'm' in self.columns:
            self._mcol = 'm'
        elif 'mass' in self.columns:
            self._mcol = 'mass'

        # Look for the keyword 'rho' or 'density' in the data.
        if 'rho' in self.columns:
            self._rhocol = 'rho'
        elif 'density' in self.columns:
            self._rhocol = 'density'

    def create_mass_column(self):
        """
        Create a new column 'm', copied from the 'massoftype' dataset parameter.

        Intended for use with Phantom data dumps.

        Raises
        ------
        KeyError
            If the 'massoftype' column does not exist in `params`.
        """
        if 'mass' not in self.params:
            raise KeyError("'mass' value does not exist in this SarracenDataFrame.")

        self['m'] = self.params['mass']
        self._mcol = 'm'

    def calc_density(self):
        """
        Create a new column 'rho' that contains particle densities.

        Density for each particle is calculated according to

            .. math::

                \\rho = m \\left( \\frac{h_{\\rm fact}}{h} \\right)^{n_{\\rm dim}}

        where :math:`m` is the particle mass, :math:`h` is the smoothing length, and :math:`h_{\\rm fact}` defines the ratio of smoothing length to particle spacing. Smoothing lengths are taken from the smoothing length column, particle masses from the mass column if present, or params if not, and hfact from params.

        Raises
        ------
        KeyError
            If the `hcol` column does not exist, there is no `mcol` column or `mass` in params, or if `hfact` does not exist in `params`.
        """
        if not {self.hcol}.issubset(self.columns):
            raise KeyError('Missing smoothing length data in this SarracenDataFrame')
        if 'hfact' not in self.params:
            raise KeyError('hfact missing from params in this SarracenDataFrame.')
        if not {self.mcol}.issubset(self.columns) and 'mass' not in self.params:
            raise KeyError('Missing particle mass data in this SarracenDataFrame.')

        mass = self.params['mass']
        # prioritize using mass per particle, if present
        if {self.mcol}.issubset(self.columns):
            mass = self[self._mcol]

        self['rho'] = (self.params['hfact'] / self['h']) ** (self.get_dim()) * mass
        self._rhocol = 'rho'

    @_copy_doc(render)
    def render(self, target: str, x: str = None, y: str = None, z: str = None, xsec: float = None,
               kernel: BaseKernel = None, x_pixels: int = None, y_pixels: int = None, xlim: Tuple[float, float] = None,
               ylim: Tuple[float, float] = None, cmap: Union[str, Colormap] = 'gist_heat', cbar: bool = True,
               cbar_kws: dict = {}, cbar_ax: Axes = None, ax: Axes = None, exact: bool = None, backend: str = None,
               integral_samples: int = 1000, rotation: np.ndarray = None, rot_origin: np.ndarray = None,
               log_scale: bool = None, dens_weight: bool = None, normalize: bool = False, hmin: bool = False,
               **kwargs) -> Axes:
        return render(self, target, x, y, z, xsec, kernel, x_pixels, y_pixels, xlim, ylim, cmap, cbar, cbar_kws,
                      cbar_ax, ax, exact, backend, integral_samples, rotation, rot_origin, log_scale, dens_weight,
                      normalize, hmin, **kwargs)

    @_copy_doc(lineplot)
    def lineplot(self, target: str, x: str = None, y: str = None, z: str = None,
                 kernel: BaseKernel = None, pixels: int = 512, xlim: Tuple[float, float] = None,
                 ylim: Tuple[float, float] = None, zlim: Tuple[float, float] = None, ax: Axes = None,
                 backend: str = None, log_scale: bool = False, dens_weight: bool = None, normalize: bool = False,
                 hmin: bool = False, **kwargs):
        return lineplot(self, target, x, y, z, kernel, pixels, xlim, ylim, zlim, ax, backend, log_scale, dens_weight,
                        normalize, hmin, **kwargs)

    @_copy_doc(streamlines)
    def streamlines(self, target: Union[Tuple[str, str], Tuple[str, str, str]], x: str = None, y: str = None,
                    z: str = None, xsec: int = None, kernel: BaseKernel = None, integral_samples: int = 1000,
                    rotation: np.ndarray = None, rot_origin: np.ndarray = None, x_pixels: int = None,
                    y_pixels: int = None, xlim: Tuple[float, float] = None, ylim: Tuple[float, float] = None,
                    ax: Axes = None, exact: bool = None, backend: str = None, dens_weight: bool = False,
                    normalize: bool = False, hmin: bool = False, **kwargs) -> Axes:
        return streamlines(self, target, x, y, z, xsec, kernel, integral_samples, rotation, rot_origin, x_pixels,
                           y_pixels, xlim, ylim, ax, exact, backend, dens_weight, normalize, hmin, **kwargs)

    @_copy_doc(arrowplot)
    def arrowplot(self, target: Union[Tuple[str, str], Tuple[str, str, str]], x: str = None, y: str = None,
                  z: str = None, xsec: int = None, kernel: BaseKernel = None, integral_samples: int = 1000,
                  rotation: np.ndarray = None, rot_origin: np.ndarray = None, x_arrows: int = None,
                  y_arrows: int = None, xlim: Tuple[float, float] = None, ylim: Tuple[float, float] = None,
                  ax: Axes = None, qkey: bool = True, qkey_kws: dict = None, exact: bool = None, backend: str = None,
                  dens_weight: bool = None, normalize: bool = False, hmin: bool = False, **kwargs) -> Axes:
        return arrowplot(self, target, x, y, z, xsec, kernel, integral_samples, rotation, rot_origin, x_arrows,
                         y_arrows, xlim, ylim, ax, qkey, qkey_kws, exact, backend, dens_weight, normalize, hmin,
                         **kwargs)

    def sph_interpolate(self, target: str, x: str = None, y: str = None, z: str = None, kernel: BaseKernel = None,
                        rotation: np.ndarray = None, rot_origin: np.ndarray = None, x_pixels: int = None,
                        y_pixels: int = None, z_pixels: int = None, xlim: Tuple[float, float] = None,
                        ylim: Tuple[float, float] = None, zlim: Tuple[float, float] = None,
                        exact: bool = None, backend: str = 'cpu', dens_weight: bool = False,
                        normalize: bool = False, hmin: bool = False) -> np.ndarray:
        """
        Interpolate this data to a 2D or 3D grid, depending on the dimensionality of the data.

        Parameters
        ----------
        target: str
            The column label of the target data.
        x, y, z: str
            The column labels of the directional data to interpolate over. Defaults to the x, y, and z columns
            detected in `data`.
        kernel: BaseKernel
            The kernel to use for smoothing the target data. Defaults to the kernel specified in `data`.
        rotation: array_like or Rotation, optional
            The rotation to apply to the data before interpolation. If defined as an array, the
            order of rotations is [z, y, x] in degrees. Only applicable to 3D datasets.
        rot_origin: array_like, optional
            Point of rotation of the data, in [x, y, z] form. Defaults to the centre
            point of the bounds of the data. Only applicable to 3D datasets.
        x_pixels, y_pixels, z_pixels: int, optional
            Number of pixels in the output image in the x, y & z directions. Default values are chosen to keep
            a consistent aspect ratio.
        xlim, ylim, zlim: tuple of float, optional
            The minimum and maximum values to use in interpolation, in particle data space. Defaults
            to the minimum and maximum values of `x`, `y` and `z`.
        exact: bool
            Whether to use exact interpolation of the data. Only applies to 2D datasets.
        backend: ['cpu', 'gpu']
            The computation backend to use when interpolating this data. Defaults to the backend specified in `data`.
        dens_weight: bool
            If True, the target will be multiplied by density. Defaults to False.
        normalize: bool
            If True, will normalize the interpolation. Defaults to False (this may change in future versions).
        hmin: bool
            If True, a minimum smoothing length of 0.5 * pixel size will be imposed. This ensures each particle
            contributes to at least one grid cell / pixel. Defaults to False (this may change in a future verison).

        Returns
        -------
        ndarray (n-Dimensional)
            The interpolated output image, in a multi-dimensional numpy array. The number of dimensions match the
            dimensions of the data. Dimensions are structured in reverse order, where (x, y, z) -> [z, y, x].

        Raises
        -------
        ValueError
            If `x_pixels`, `y_pixels` or `z_pixels` are less than or equal to zero, or
            if the specified `x`, `y` and `z` minimum and maximum values result in an invalid region, or
            if `data` is not 2 or 3 dimensional.
        KeyError
            If `target`, `x`, `y`, `z`, mass, density, or smoothing length columns do not
            exist in `data`.
        """
        if self.get_dim() == 2:
            if xlim is None:
                xlim = (None, None)
            if ylim is None:
                ylim = (None, None)
            return interpolate_2d(self, target, x, y, kernel, x_pixels, y_pixels, xlim, ylim, exact, backend,
                                  dens_weight, normalize, hmin)
        elif self.get_dim() == 3:
            return interpolate_3d_grid(self, target, x, y, z, kernel, rotation, rot_origin, x_pixels, y_pixels,
                                       z_pixels, xlim, ylim, zlim, backend, dens_weight, normalize, hmin)

    @property
    def params(self):
        """
        dict: Miscellaneous dataset-level parameters.

        Raises
        ------
        TypeError
            If `params` is set to a non-dictionary or non-None object.
        """
        return self._params

    @params.setter
    def params(self, new_params):
        if new_params is None:
            self._params = None
            return
        if not type(new_params) is dict:
            raise TypeError("Parameters not a dictionary")
        self._params = new_params

    @property
    def units(self):
        """Series: Units for each column of this dataset."""
        return self._units

    @units.setter
    def units(self, new_units: Series):
        self._units = new_units

    @property
    def xcol(self):
        """
        str : Label of the column which contains x-positional data.

        If this is set to a column which does not exist in the dataset, the column
        label will remain set to the old value.
        """
        return self._xcol

    @xcol.setter
    def xcol(self, new_col: str):
        if new_col in self:
            self._xcol = new_col

    @property
    def ycol(self):
        """
        str : Label of the column which contains y-positional data.

        If this is set to a column which does not exist in the dataset, the column
        label will remain set to the old value.
        """
        return self._ycol

    @ycol.setter
    def ycol(self, new_col: str):
        if new_col in self:
            self._ycol = new_col

    @property
    def zcol(self):
        """
        str : Label of the column which contains z-positional data.

        If this is set to a column which does not exist in the dataset, the column
        label will remain set to the old value.
        """
        return self._zcol

    @zcol.setter
    def zcol(self, new_col: str):
        if new_col in self:
            self._zcol = new_col

    @property
    def hcol(self):
        """
        str : Label of the column which contains smoothing length data.

        If this is set to a column which does not exist in the dataset, the column
        label will remain set to the old value.
        """
        return self._hcol

    @hcol.setter
    def hcol(self, new_col: str):
        if new_col in self:
            self._hcol = new_col

    @property
    def mcol(self):
        """
        str : Label of the column which contains particle mass data.

        If this is set to a column which does not exist in the dataset, the column
        label will remain set to the old value.
        """
        return self._mcol

    @mcol.setter
    def mcol(self, new_col: str):
        if new_col in self:
            self._mcol = new_col

    @property
    def rhocol(self):
        """
        str : Label of the column which contains particle density data.

        If this is set to a column which does not exist in the dataset, the column
        label will remain set to the old value.
        """
        return self._rhocol

    @rhocol.setter
    def rhocol(self, new_col: str):
        if new_col in self:
            self._rhocol = new_col

    @property
    def kernel(self):
        """
        BaseKernel : The default kernel to use for interpolation operations with this dataset.

        If this is set to an object which is not a BaseKernel, the kernel will remain set as
        the old value.
        """
        return self._kernel

    @kernel.setter
    def kernel(self, new_kernel: BaseKernel):
        if isinstance(new_kernel, BaseKernel):
            self._kernel = new_kernel

    @property
    def backend(self):
        """
        ['cpu', 'gpu'] : The default backend to use for interpolation operations with this dataset.

        'cpu' - Best for small datasets, or cases where a GPU is not available.
        'gpu' - Best for large datasets, with a CUDA-enabled GPU.
        """
        return self._backend

    @backend.setter
    def backend(self, new_backend: str):
        self._backend = new_backend

    def get_dim(self):
        """
        Get the dimensionality of the data in this dataframe.

        Returns
        -------
        int
            The number of positional dimensions.
        """
        return 3 if self._zcol is not None else 2



"""pytest unit tests for render.py functions."""
import pandas as pd
from matplotlib import pyplot as plt
from numba import cuda
from numpy.testing import assert_array_equal
from pytest import mark

from sarracen import SarracenDataFrame, interpolate_2d, interpolate_2d_line, interpolate_3d_proj, interpolate_3d_cross
from sarracen.render import render, streamlines, arrowplot, lineplot

backends = ['cpu']
if cuda.is_available():
    backends.append('gpu')

@mark.parametrize("backend", backends)
def test_interpolation_passthrough(backend):
    """
    Verify that each rendering function uses the proper underlying interpolation function.
    """
    df = pd.DataFrame({'x': [3, 6], 'y': [5, 1], 'P': [1, 1], 'Ax': [3, 2], 'Ay': [2, 1], 'h': [1, 1], 'rho': [1, 1],
                       'm': [1, 1]})
    sdf = SarracenDataFrame(df)
    sdf.backend = backend

    fig, ax = plt.subplots()
    render(sdf, 'P', ax=ax)
    assert_array_equal(ax.images[0].get_array(), interpolate_2d(sdf, 'P'))
    plt.close(fig)

    fig, ax = plt.subplots()
    lineplot(sdf, 'P', xlim=(3, 6), ylim=(1, 5), ax=ax)
    assert_array_equal(ax.lines[0].get_ydata(), interpolate_2d_line(sdf, 'P', xlim=(3, 6), ylim=(1, 5)))
    plt.close(fig)

    df = pd.DataFrame({'x': [3, 6], 'y': [5, 1], 'z': [2, 1], 'P': [1, 1], 'h': [1, 1], 'rho': [1, 1], 'm': [1, 1]})
    sdf = SarracenDataFrame(df)
    sdf.backend = backend

    fig, ax = plt.subplots()
    render(sdf, 'P', ax=ax)
    assert_array_equal(ax.images[0].get_array(), interpolate_3d_proj(sdf, 'P'))
    plt.close(fig)

    fig, ax = plt.subplots()
    render(sdf, 'P', xsec=1.5, ax=ax)
    assert_array_equal(ax.images[0].get_array(), interpolate_3d_cross(sdf, 'P'))
    plt.close(fig)


@mark.parametrize("backend", backends)
def test_cmap(backend):
    """
    Verify that each rendering function uses the provided color map.
    """
    df = pd.DataFrame({'x': [3, 6], 'y': [5, 1], 'P': [1, 1], 'h': [1, 1], 'rho': [1, 1], 'm': [1, 1]})
    sdf = SarracenDataFrame(df)
    sdf.backend = backend

    fig, ax = plt.subplots()
    render(sdf, 'P', cmap='magma', ax=ax)
    assert ax.images[0].cmap.name == 'magma'
    plt.close(fig)

    df = pd.DataFrame({'x': [3, 6], 'y': [5, 1], 'z': [2, 1], 'P': [1, 1], 'h': [1, 1], 'rho': [1, 1], 'm': [1, 1]})
    sdf = SarracenDataFrame(df)
    sdf.backend = backend

    fig, ax = plt.subplots()
    render(sdf, 'P', cmap='magma', ax=ax)
    assert ax.images[0].cmap.name == 'magma'
    plt.close(fig)

    fig, ax = plt.subplots()
    render(sdf, 'P', xsec=1.5, cmap='magma', ax=ax)
    assert ax.images[0].cmap.name == 'magma'
    plt.close(fig)


@mark.parametrize("backend", backends)
def test_cbar_exclusion(backend):
    """
    Verify that each rendering function respects the cbar argument.
    """
    df_2 = pd.DataFrame({'x': [3, 6], 'y': [5, 1], 'P': [1, 1], 'h': [1, 1], 'rho': [1, 1], 'm': [1, 1]})
    sdf_2 = SarracenDataFrame(df_2)
    sdf_2.backend = backend
    df_3 = pd.DataFrame({'x': [3, 6], 'y': [5, 1], 'z': [2, 1], 'P': [1, 1], 'h': [1, 1], 'rho': [1, 1], 'm': [1, 1]})
    sdf_3 = SarracenDataFrame(df_3)
    sdf_3.backend = backend

    for args in [{'data': sdf_2, 'xsec': None}, {'data': sdf_3, 'xsec': None}, {'data': sdf_3, 'xsec': 1.5}]:
        fig, ax = plt.subplots()
        render(args['data'], 'P', xsec=args['xsec'], cbar=True, ax=ax)
        assert ax.images[-1].colorbar is not None
        plt.close(fig)

        fig, ax = plt.subplots()
        render(args['data'], 'P', xsec=args['xsec'], cbar=False, ax=ax)
        assert ax.images[-1].colorbar is None
        plt.close(fig)


@mark.parametrize("backend", backends)
def test_cbar_keywords(backend):
    """
    Verify that each rendering function respects the passed keywords for the colorbar.
    """
    df_2 = pd.DataFrame({'x': [3, 6], 'y': [5, 1], 'P': [1, 1], 'h': [1, 1], 'rho': [1, 1], 'm': [1, 1]})
    sdf_2 = SarracenDataFrame(df_2)
    sdf_2.backend = backend
    df_3 = pd.DataFrame({'x': [3, 6], 'y': [5, 1], 'z': [2, 1], 'P': [1, 1], 'h': [1, 1], 'rho': [1, 1], 'm': [1, 1]})
    sdf_3 = SarracenDataFrame(df_3)
    sdf_3.backend = backend

    for args in [{'data': sdf_2, 'xsec': None}, {'data': sdf_3, 'xsec': None}, {'data': sdf_3, 'xsec': 1.5}]:
        fig, ax = plt.subplots()
        render(args['data'], 'P', xsec=args['xsec'], cbar_kws={'orientation': 'horizontal'}, ax=ax)
        assert ax.images[-1].colorbar.orientation == 'horizontal'
        plt.close(fig)


@mark.parametrize("backend", backends)
def test_kwargs(backend):
    """
    Verify that each rendering function respects passed keyword arguments.
    """
    df_2 = pd.DataFrame({'x': [3, 6], 'y': [5, 1], 'P': [1, 1], 'h': [1, 1], 'rho': [1, 1], 'm': [1, 1], 'Ax': [1, 1],
                         'Ay': [1, 1]})
    sdf_2 = SarracenDataFrame(df_2)
    sdf_2.backend = backend
    df_3 = pd.DataFrame({'x': [3, 6], 'y': [5, 1], 'z': [2, 1], 'P': [1, 1], 'h': [1, 1], 'rho': [1, 1], 'm': [1, 1]})
    sdf_3 = SarracenDataFrame(df_3)
    sdf_3.backend = backend

    for args in [{'data': sdf_2, 'xsec': None}, {'data': sdf_3, 'xsec': None}, {'data': sdf_3, 'xsec': 1.5}]:
        fig, ax = plt.subplots()
        render(args['data'], 'P', xsec=args['xsec'], ax=ax, origin='upper')
        assert ax.images[0].origin == 'upper'
        plt.close(fig)

    fig, ax = plt.subplots()
    streamlines(sdf_2, ('Ax', 'Ay'), ax=ax, zorder=5)
    assert ax.patches[0].zorder == 5
    plt.close(fig)

    fig, ax = plt.subplots()
    arrowplot(sdf_2, ('Ax', 'Ay'), ax=ax, zorder=5)
    assert ax.collections[0].zorder == 5
    plt.close(fig)

    fig, ax = plt.subplots()
    lineplot(sdf_2, 'P', xlim=(3, 6), ylim=(1, 5), ax=ax, linestyle='--')
    assert ax.lines[0].get_linestyle() == '--'
    plt.close(fig)


@mark.parametrize("backend", backends)
def test_rotated_ticks(backend):
    """
    A rotated plot should have no x & y ticks.
    """
    df = pd.DataFrame({'x': [3, 6], 'y': [5, 1], 'z': [2, 1], 'P': [1, 1], 'h': [1, 1], 'rho': [1, 1], 'm': [1, 1],
                       'Ax': [1, 1], 'Ay': [1, 1], 'Az': [1, 1]})
    sdf = SarracenDataFrame(df)
    sdf.backend = backend

    for xsec in [None, 1.5]:
        fig, ax = plt.subplots()
        render(sdf, 'P', xsec=xsec, ax=ax, rotation=[34, 23, 50])

        assert ax.get_xticks().size == 0
        assert ax.get_yticks().size == 0
        plt.close(fig)


    for func in [arrowplot, streamlines]:
        fig, ax = plt.subplots()
        func(sdf, ('Ax', 'Ay', 'Az'), rotation=[34, 23, 50], ax=ax)

        assert ax.get_xticks().size == 0
        assert ax.get_yticks().size == 0
        plt.close(fig)



@mark.parametrize("backend", backends)
def test_plot_labels(backend):
    """
    Verify that plot labels for each rendering function are correct.
    """
    df_2 = pd.DataFrame({'x': [3, 6], 'y': [5, 1], 'P': [1, 1], 'h': [1, 1], 'rho': [1, 1], 'm': [1, 1], 'Ax': [1, 1],
                         'Ay': [1, 1]})
    sdf_2 = SarracenDataFrame(df_2)
    sdf_2.backend = backend

    df_3 = pd.DataFrame({'x': [3, 6], 'y': [5, 1], 'z': [2, 1], 'P': [1, 1], 'h': [1, 1], 'rho': [1, 1], 'm': [1, 1],
                         'Ax': [1, 1], 'Ay': [1, 1], 'Az': [1, 1]})
    sdf_3 = SarracenDataFrame(df_3)
    sdf_3.backend = backend

    for args in [{'data': sdf_2, 'xsec': None}, {'data': sdf_3, 'xsec': None}, {'data': sdf_3, 'xsec': 0}]:
        fig, ax = plt.subplots()
        render(args['data'], 'P', xsec=args['xsec'], ax=ax)

        assert ax.get_xlabel() == 'x'
        assert ax.get_ylabel() == 'y'
        assert ax.figure.axes[1].get_ylabel() == \
               ('column ' if args['data'] is sdf_3 and args['xsec'] is None else '') + 'P'
        plt.close(fig)

        fig, ax = plt.subplots()
        render(args['data'], 'rho', x='y', y='x', xsec=args['xsec'], ax=ax)

        assert ax.get_xlabel() == 'y'
        assert ax.get_ylabel() == 'x'
        assert ax.figure.axes[1].get_ylabel() == ('column ' if args['data'] is sdf_3 and args['xsec'] is None else '')\
               + 'rho'
        plt.close(fig)

    for func in [streamlines, arrowplot]:
        fig, ax = plt.subplots()
        func(sdf_2, ('Ax', 'Ay'), ax=ax)

        assert ax.get_xlabel() == 'x'
        assert ax.get_ylabel() == 'y'
        plt.close(fig)

        fig, ax = plt.subplots()
        func(sdf_3, ('Ax', 'Ay', 'Az'), x='y', y='x', ax=ax)

        assert ax.get_xlabel() == 'y'
        assert ax.get_ylabel() == 'x'
        plt.close(fig)

    fig, ax = plt.subplots()
    lineplot(sdf_2, 'P', ax=ax)

    assert ax.get_xlabel() == 'cross-section (x, y)'
    assert ax.get_ylabel() == 'P'
    plt.close(fig)

    fig, ax = plt.subplots()
    lineplot(sdf_2, 'rho', x='y', y='x', ax=ax)

    assert ax.get_xlabel() == 'cross-section (y, x)'
    assert ax.get_ylabel() == 'rho'
    plt.close(fig)


@mark.parametrize("backend", backends)
def test_plot_bounds(backend):
    """
    Verify that plot bounds are set correctly for each rendering function.
    """
    df_2 = pd.DataFrame({'x': [6, 3], 'y': [5, 1], 'P': [1, 1], 'h': [1, 1], 'rho': [1, 1], 'm': [1, 1]})
    sdf_2 = SarracenDataFrame(df_2)
    sdf_2.backend = backend

    df_3 = pd.DataFrame({'x': [6, 3], 'y': [5, 1], 'z': [2, 1], 'P': [1, 1], 'h': [1, 1], 'rho': [1, 1], 'm': [1, 1],
                         'Ax': [1, 1], 'Ay': [1, 1], 'Az': [1, 1]})
    sdf_3 = SarracenDataFrame(df_3)
    sdf_3.backend = backend

    for args in [{'data': sdf_2, 'xsec': None}, {'data': sdf_3, 'xsec': None}, {'data': sdf_3, 'xsec': 1.5}]:
        fig, ax = plt.subplots()
        render(args['data'], 'P', xsec=args['xsec'], ax=ax)

        assert ax.get_xlim() == (3, 6)
        assert ax.get_ylim() == (1, 5)

        if args['data'] is sdf_2:
            assert ax.figure.axes[1].get_ylim() == (0, interpolate_2d(sdf_2, 'P').max())
        else:
            if args['xsec']:
                assert ax.figure.axes[1].get_ylim() == (0, interpolate_3d_cross(sdf_3, 'P').max())
            else:
                assert ax.figure.axes[1].get_ylim() == (0, interpolate_3d_proj(sdf_3, 'P').max())
        plt.close(fig)

    fig, ax = plt.subplots()
    lineplot(sdf_2, 'P', ax=ax)

    assert ax.get_xlim() == (0, 5)
    # 512 pixels across (by default), and both particles are in corners
    # therefore closest pixel to a particle is sqrt(41)/1024 units away
    # use default kernel to determine the max pressure value
    assert ax.get_ylim() == (0, interpolate_2d_line(sdf_2, 'P').max())
    plt.close(fig)

    for func in [arrowplot, streamlines]:
        fig, ax = plt.subplots()
        func(sdf_3, ('Ax', 'Ay', 'Az'), ax=ax)

        assert ax.get_xlim() == (3, 6)
        assert ax.get_ylim() == (1, 5)
        plt.close(fig)


@mark.parametrize("backend", backends)
def test_snap(backend):
    df = pd.DataFrame({'x': [0.0001, 5.2],
                       'y': [3.00004, 0.1],
                       'P': [1, 1],
                       'h': [1, 1],
                       'rho': [1, 1],
                       'm': [1, 1]})
    sdf = SarracenDataFrame(df)
    sdf.backend = backend

    fig, ax = plt.subplots()
    sdf.render('P', ax=ax)

    # 0.0001 -> 0.0, 5.2 -> 5.2
    assert ax.get_xlim() == (0.0, 5.2)
    # 0.1 -> 0.1, 3.00004 -> 3.0
    assert ax.get_ylim() == (0.1, 3.0)
    plt.close(fig)
